import unittest
from os.path import abspath, dirname, join

import pandas as pd
import wntr
from pandas.testing import assert_frame_equal, assert_series_equal

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestSegmentation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        inp_file = join(ex_datadir, "Net3.inp")
        self.wn = wntr.network.WaterNetworkModel(inp_file)
        G = self.wn.to_graph()
        self.valves = pd.read_csv(
            join(test_datadir, "valve_layer_stategic_1.csv"),
            index_col=0,
            dtype="object",
        )
        (
            self.node_segments,
            self.link_segments,
            self.seg_size,
        ) = wntr.metrics.topographic.valve_segments(G, self.valves)

        # import matplotlib
        # cmap = matplotlib.colors.ListedColormap(np.random.rand(num_segments,3))
        # wntr.graphics.plot_network(self.wn2, node_segments, link_segments, valve_layer=valves,
        #                            node_cmap=cmap, link_cmap=cmap,
        #                            node_range=[0.5,num_segments+0.5],
        #                            link_range=[0.5,num_segments+0.5])

    @classmethod
    def tearDownClass(self):
        pass

    def test_valve_criticality_length(self):

        # Gather the link lengths for the length-based criticality calculation
        link_lengths = self.wn.query_link_attribute("length")

        # Calculate the length-based valve criticality for each valve
        valve_crit = wntr.metrics.topographic._valve_criticality_length(
            link_lengths, self.valves, self.node_segments, self.link_segments
        )

        # valve_crit.to_csv('valve_crit_length.csv')
        # filename = 'valve_crit_length.jpg'
        # wntr.graphics.plot_network(self.wn, valve_layer=self.valves,
        #                            valve_criticality=valve_crit,
        #                            node_size=10, filename=filename)

        expected_valve_crit = pd.read_csv(
            join(test_datadir, "valve_crit_length.csv"), index_col=[0], 
        ).squeeze("columns")

        assert_series_equal(
            valve_crit, expected_valve_crit, check_dtype=False, check_names=False
        )

    def test_valve_criticality_demand(self):

        # Gather the node demands for the demand-based criticality calculation
        node_demands = wntr.metrics.average_expected_demand(self.wn)

        # Calculate the demand-based valve criticality for each valve
        valve_crit = wntr.metrics.topographic._valve_criticality_demand(
            node_demands, self.valves, self.node_segments, self.link_segments
        )

        # valve_crit.to_csv('valve_crit_demand.csv')
        # filename = 'valve_crit_demand.jpg'
        # wntr.graphics.plot_network(self.wn, valve_layer=self.valves,
        #                            valve_criticality=valve_crit,
        #                            node_size=10, filename=filename)

        expected_valve_crit = pd.read_csv(
            join(test_datadir, "valve_crit_demand.csv"), index_col=[0], 
        ).squeeze("columns")

        assert_series_equal(
            valve_crit, expected_valve_crit, check_dtype=False, check_names=False
        )

    def test_valve_criticality(self):

        # Calculate the valve-based valve criticality for each valve
        valve_crit = wntr.metrics.topographic._valve_criticality(
            self.valves, self.node_segments, self.link_segments
        )

        # valve_crit.to_csv('valve_crit_valve.csv')
        # filename = 'valve_crit_valve.jpg'
        # wntr.graphics.plot_network(self.wn, valve_layer=self.valves,
        #                            valve_criticality=valve_crit,
        #                            node_size=10, filename=filename)

        expected_valve_crit = pd.read_csv(
            join(test_datadir, "valve_crit_valve.csv"), index_col=[0], 
        ).squeeze("columns")

        assert_series_equal(
            valve_crit, expected_valve_crit, check_dtype=False, check_names=False
        )


if __name__ == "__main__":
    unittest.main()




def _check_float(value, property_name: str) -> float:
    """Transform a value to a float.

    Raises ValueError if the value is not a float or convertible to float.
    """
    try:
        return float(value)
    except (ValueError, TypeError) as e:
        value_type = type(value).__name__
        raise ValueError(
            f"{property_name} must be a float or convertible to float. Received {value} of type {value_type}"
        ) from e


def _check_positive_or_zero_float(value, property_name: str) -> float:
    """Transform a value to a float and check it is positive or zero.

    Raises ValueError if the value is not a float or convertible to float,
    or if the value is negative.
    """
    value = _check_float(value, property_name)

    if value < 0:
        raise ValueError(f"{property_name} must not be negative")

    return value


def _check_positive_non_zero_float(value, property_name: str) -> float:
    """Transform a value to a float and check it is positive.

    Raises ValueError if the value is not a float or convertible to float,
    or if the value is not positive.
    """
    value = _check_float(value, property_name)

    if not value > 0:
        raise ValueError(f"{property_name} must be greater than zero")

    return value


def _check_float_or_none(value, property_name: str) -> float | None:
    """Transform a value to a float, unless it is None in which case return None.

    Raises ValueError if the value is not a float or convertible to float,
    and is not None.
    """
    if value is None:
        return None

    try:
        return float(value)
    except (ValueError, TypeError) as e:
        value_type = type(value).__name__
        raise ValueError(
            f"{property_name} must be a float, convertible to float, or None. Received {value} of type {value_type}"
        ) from e



import os
import sys
import unittest
import warnings
from os.path import abspath, dirname, isfile, join

import networkx as nx
import matplotlib.pylab as plt
import matplotlib
from wntr.graphics.color import custom_colormap
import pandas as pd
import numpy as np
import wntr

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestGraphics(unittest.TestCase):
    def test_plot_network1(self):
        filename = abspath(join(testdir, "plot_network1.png"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net6.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        wntr.graphics.plot_network(wn)
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

    def test_plot_network2(self):
        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        # undirected
        filename = abspath(join(testdir, "plot_network2_undirected.png"))
        if isfile(filename):
            os.remove(filename)

        wntr.graphics.plot_network(
            wn, node_attribute="elevation", link_attribute="length"
        )
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))
        
        # directed
        filename = abspath(join(testdir, "plot_network2_directed.png"))
        if isfile(filename):
            os.remove(filename)

        wntr.graphics.plot_network(
            wn, node_attribute="elevation", link_attribute="length", directed=True
        )
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

    def test_plot_network3(self):
        filename = abspath(join(testdir, "plot_network3.png"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        wntr.graphics.plot_network(
            wn,
            node_attribute=["11", "21"],
            link_attribute=["112", "113"],
            link_labels=True,
        )
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

    def test_plot_network4(self):
        filename = abspath(join(testdir, "plot_network4.png"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        wntr.graphics.plot_network(
            wn,
            node_attribute={"11": 5, "21": 10},
            link_attribute={"112": 3, "113": 9},
            node_labels=True,
        )
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

    def test_plot_network5(self):
        filename = abspath(join(testdir, "plot_network5.png"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        pop = wntr.metrics.population(wn)

        wntr.graphics.plot_network(
            wn, node_attribute=pop, node_range=[0, 500], title="Population"
        )
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

    def test_plot_network6(self):
        # legend
        filename = abspath(join(testdir, "plot_network6.png"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net6.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        wntr.graphics.plot_network(
            wn, node_attribute="elevation", link_attribute="diameter", 
            add_colorbar=True, legend=True
        )
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))
    
    def test_plot_network_options(self):
        # NOTE:to compare with the old plot_network set compare=True.
        #   this should be set to false for regular testing
        compare = False
        
        cmap = matplotlib.colormaps['viridis']
        
        inp_file = join(ex_datadir, "Net6.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        
        random_node_values = pd.Series(
            np.random.rand(len(wn.node_name_list)), index=wn.node_name_list)
        random_link_values = pd.Series(
            np.random.rand(len(wn.link_name_list)), index=wn.link_name_list)
        random_pipe_values = pd.Series(
            np.random.rand(len(wn.pipe_name_list)), index=wn.pipe_name_list)
        random_node_dict_subset = dict(random_node_values.iloc[:10])
        random_link_dict_subset = dict(random_link_values.iloc[:10])
        node_list = list(wn.node_name_list[:10])
        link_list = list(wn.link_name_list[:10])
        
        kwarg_list = [
            {"node_attribute": "elevation",
             "node_range": [0,20],
             "node_alpha": 0.5,
             "node_colorbar_label": "test_label"},
            {"link_attribute": "diameter",
             "link_range": [0,None],
             "link_alpha": 0.5,
             "link_colorbar_label": "test_label"},
            {"link_attribute": "diameter",
            "node_attribute": "elevation"},
            {"node_labels": True,
             "link_labels": True},
            {"node_attribute": "elevation",
             "add_colorbar": False},
            {"link_attribute": "diameter",
             "add_colorbar": False},
            {"node_attribute": node_list},
            {"node_attribute": random_node_values},
            {"node_attribute": random_node_dict_subset},
            {"link_attribute": link_list},
            {"link_attribute": random_link_values},
            {"link_attribute": random_link_dict_subset},
            {"directed": True},
            {"link_attribute": random_pipe_values,
             "node_size": 0,
             "link_cmap": cmap,
             "link_range": [0,1],
             "link_width": 1.5},
        ]
        
        for kwargs in kwarg_list:
            filename = abspath(join(testdir, "plot_network_options.png"))
            if isfile(filename):
                os.remove(filename)
            if compare:
                fig, ax = plt.subplots(1,2)
                wntr.graphics.plot_network(wn, ax=ax[0], title="GIS plot_network", backend='gpd', **kwargs)
                wntr.graphics.plot_network(wn, ax=ax[1], title="NX plot_network", backend='nx', **kwargs)
                fig.savefig(filename, format="png")
                plt.close(fig)
            else:
                wntr.graphics.plot_network(wn, backend='nx', **kwargs)
                plt.savefig(filename, format="png")
                plt.close()
                
            self.assertTrue(isfile(filename))
            os.remove(filename)
            

    def test_plot_interactive_network1(self):

        filename = abspath(join(testdir, "plot_interactive_network1.html"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        plt.figure()
        wntr.graphics.plot_interactive_network(
            wn, node_attribute=["107", "123"], filename=filename, auto_open=False
        )

        self.assertTrue(isfile(filename))

    def test_plot_leaflet_network1(self):

        filename = abspath(join(testdir, "plot_leaflet_network1.html"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        longlat_map = {"Lake": (-106.6587, 35.0623), "219": (-106.5248, 35.1918)}
        wn2 = wntr.morph.convert_node_coordinates_to_longlat(wn, longlat_map)

        plt.figure()
        wntr.graphics.plot_leaflet_network(
            wn2,
            node_attribute="elevation",
            link_attribute="length",
            add_legend=True,
            filename=filename,
        )

        self.assertTrue(isfile(filename))

    def test_network_animation1(self):
        filename = abspath(join(testdir, "plot_leaflet_network1.html"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim()

        pressure = results.node["pressure"]
        flowrate = results.link["flowrate"]
        anim = wntr.graphics.network_animation(
            wn, node_attribute=pressure, link_attribute=flowrate, repeat=True
        )

        from matplotlib.animation import FuncAnimation

        self.assertTrue(isinstance(anim, FuncAnimation))

    def test_plot_fragility_curve1(self):
        from scipy.stats import lognorm

        filename = abspath(join(testdir, "plot_fragility_curve1.png"))
        if isfile(filename):
            os.remove(filename)

        FC = wntr.scenario.FragilityCurve()
        FC.add_state("Minor", 1, {"Default": lognorm(0.5, scale=0.3)})
        FC.add_state("Major", 2, {"Default": lognorm(0.5, scale=0.7)})

        plt.figure()
        wntr.graphics.plot_fragility_curve(FC)
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

    def test_plot_pump_curve1(self):
        filename = abspath(join(testdir, "plot_pump_curve1.png"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        pump = wn.get_link("10")

        plt.figure()
        wntr.graphics.plot_pump_curve(pump)
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

    def test_plot_tank_curve(self):
        filename = abspath(join(testdir, "plot_tank_curve.png"))
        if isfile(filename):
            os.remove(filename)

        inp_file = join(test_datadir, "Anytown_multipointcurves.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        tank_w_curve = wn.get_node("41")
        tank_no_curve = wn.get_node("42")

        plt.figure()
        shouldBeAxis = wntr.graphics.plot_tank_volume_curve(tank_w_curve)
        plt.savefig(filename, format="png")
        plt.close()

        self.assertTrue(isfile(filename))

        shouldBeNone = wntr.graphics.plot_tank_volume_curve(tank_no_curve)
        self.assertTrue(shouldBeNone is None)

    def test_custom_colormap(self):
        cmp = wntr.graphics.custom_colormap(
            3, colors=["blue", "white", "red"], name="custom"
        )
        self.assertEqual(cmp.N, 3)
        self.assertEqual(cmp.name, "custom")
        
        


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr
import wntr.msx
import wntr.epanet.msx
import wntr.epanet.msx.toolkit

testdir = dirname(abspath(str(__file__)))
test_network_dir = join(testdir, "networks_for_testing")
inp_filename = join(test_network_dir, 'msx_example.inp')
msx_filename = join(test_network_dir, 'msx_example.msx')

class Test(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    def test_msx_io(self):
        wn_model = wntr.network.WaterNetworkModel(inp_file_name=inp_filename)
        msx_model = wntr.msx.MsxModel(msx_file_name=msx_filename)
        wntr.epanet.msx.toolkit.MSXepanet(inp_filename, msxfile=msx_filename)
        


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr
from scipy.stats import lognorm, norm

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "..", "..", "tests", "networks_for_testing")

FC1 = wntr.scenario.FragilityCurve()
FC1.add_state("Major", 2, {"Default": norm(loc=1, scale=2)})
FC1.add_state("Minor", 1, {"Default": norm(loc=0, scale=1)})

FC2 = wntr.scenario.FragilityCurve()
FC2.add_state(
    "Minor",
    1,
    {"Default": lognorm(0.25, loc=0, scale=1), "3": lognorm(0.2, loc=0, scale=1)},
)
FC2.add_state("Major", 2, {"Default": lognorm(0.25, loc=1, scale=2)})

# x = np.linspace(-5,5,100)
# for name, state in FC2.states():
#    dist=state.distribution['Default']
#    plt.plot(x,dist.cdf(x), label=name)
# plt.ylim((0,1))
# plt.legend()


class TestFragilityCurveScenario(unittest.TestCase):
    def test_get_priority_map(self):
        priority_map = FC1.get_priority_map()
        self.assertDictEqual(priority_map, {None: 0, "Minor": 1, "Major": 2})

    def test_cdf_probability(self):
        x = pd.Series({"1": 0, "2": 1, "3": 2})
        Pr = FC1.cdf_probability(x)
        self.assertEqual(Pr.loc["1", "Minor"], 0.5)
        self.assertLess(Pr.loc["2", "Minor"] - 0.841, 0.001)
        self.assertLess(Pr.loc["3", "Minor"] - 0.977, 0.001)
        self.assertEqual(Pr.loc["2", "Major"], 0.5)

    def test_sample_damage_state(self):
        x = pd.Series({"1": 0, "2": 1, "3": 2})
        Pr = FC1.cdf_probability(x)
        states = FC1.sample_damage_state(Pr, seed=45)
        # p with random seed of 45
        # 1    0.989012
        # 2    0.549545
        # 3    0.281447
        self.assertEqual(states.loc["1"], None)
        self.assertEqual(states.loc["2"], "Minor")
        self.assertEqual(states.loc["3"], "Major")


if __name__ == "__main__":
    unittest.main()



"""
The wntr.metrics.hydraulic module contains hydraulic metrics.
"""
import wntr.network
import numpy as np
import pandas as pd
import networkx as nx
import math
from collections import Counter
import sys
from functools import reduce
    
import logging

logger = logging.getLogger(__name__)

def expected_demand(wn, start_time=None, end_time=None, timestep=None, category=None):
    """
    Compute expected demand at each junction and time using base demands
    and demand patterns along with the demand multiplier
    
    Parameters
    -----------
    wn : wntr WaterNetworkModel
        Water network model. The water network model is needed to 
        get demand timeseries at junctions and options related to 
        duration, timestep, and demand multiplier.
        
    start_time : int (optional)
        Start time in seconds, if None then value is set to 0
        
    end_time : int  (optional)
        End time in seconds, if None then value is set to wn.options.time.duration

    timestep : int (optional)
        Timestep, if None then value is set to wn.options.time.report_timestep
    
    category : str (optional)
        Demand category name.  If None, all demand categories are used.
            
    Returns
    -------
    A pandas DataFrame that contains expected demand in m3/s (index = times, columns = junction names).
    """
    if start_time is None:
        start_time = 0
    if end_time is None:
        end_time = wn.options.time.duration
    if timestep is None:
        timestep = wn.options.time.report_timestep
        
    exp_demand = {}
    tsteps = np.arange(start_time, end_time+timestep, timestep)
    for name, junc in wn.junctions():
        dem = []
        for ts in tsteps:
            dem.append(junc.demand_timeseries_list.at(ts, 
                       multiplier=wn.options.hydraulic.demand_multiplier, category=category))
        exp_demand[name] = dem 
    
    exp_demand = pd.DataFrame(index=tsteps, data=exp_demand)
    
    return exp_demand

def average_expected_demand(wn, category=None):
    """
    Compute average expected demand per day at each junction using base demands
    and demand patterns along with the demand multiplier
    
    Parameters
    -----------
    wn : wntr WaterNetworkModel
        Water network model. The water network model is needed to 
        get demand timeseries at junctions and options related to 
        duration, timestep, and demand multiplier.
    
    category : str (optional)
        Demand category name.  If None, all demand categories are used.
        
    Returns
    -------
    A pandas Series that contains average expected demand in m3/s (index = junction names).
    """
    L = [24*3600] # start with a 24 hour pattern
    for name, pattern in wn.patterns():
        L.append(len(pattern.multipliers)*wn.options.time.pattern_timestep)
    lcm = int(_lcml(L))
    
    start_time = wn.options.time.pattern_start
    end_time = start_time+lcm
    timestep = wn.options.time.pattern_timestep
        
    exp_demand = expected_demand(wn, start_time, end_time-timestep, timestep, category=category)
    ave_exp_demand = exp_demand.mean(axis=0)

    return ave_exp_demand

def _gcd(x,y):
  while y:
    if y<0:
      x,y=-x,-y
    x,y=y,x % y
    return x

def _gcdl(*list):
  return reduce(_gcd, *list)

def _lcm(x,y):
  return x*y / _gcd(x,y)

def _lcml(*list):
  return reduce(_lcm, *list)

def water_service_availability(expected_demand, demand):
    r"""
    Compute water service availability (WSA) at junctions, defined as follows:
        
    .. math:: WSA = \dfrac{demand}{expected\_demand}
        
    where 
    :math:`demand` is the actual demand computed from a hydraulic simulation, and 
    :math:`expected\_demand` is the expected demand computed from base demands and demand 
    patterns. Expected demand can be computed using the 
    :class:`~wntr.metrics.hydraulic.expected_demand` method.

    WSA can be averaged over times and/or nodes (see below).  If 
    expected demand is 0 for a particular junction, water service availability 
    will be set to NaN for that junction. 

    * To compute water service availability for each junction and timestep, 
      expected_demand and demand should be pandas DataFrames (index = times, columns = junction names). 
    
    * To compute an average water service availability for each junction (averaged over time), 
      expected_demand and demand should be a pandas Series, indexed by junction.  
      To convert a DataFrame (index = times, columns = junction names) to a 
      Series indexed by junction, use the following code:
    
        :math:`expected\_demand.sum(axis=0)`
		
        :math:`demand.sum(axis=0)`
    
    * To compute an average water service availability for each timestep (averaged over junctions), 
      expected_demand and demand should be a pandas Series, indexed by time.  
      To convert a DataFrame (index = times, columns = junction names) to a 
      Series indexed by time, use the following code:
        
        :math:`expected\_demand.sum(axis=1)`
		
        :math:`demand.sum(axis=1)`
        
    Parameters
    ----------
    expected_demand : pandas DataFrame or pandas Series (see note above)
        Expected demand at junctions 
    
    demand : pandas DataFrame or pandas Series (see note above)
        Actual demand (generally from a PDD hydraulic simulation) at junctions

    Returns
    -------
    A pandas DataFrame or pandas Series that contains water service 
    availability.
    """

    wsa = demand.div(expected_demand) 
    
    return wsa

def todini_index(head, pressure, demand, flowrate, wn, Pstar):
    """
    Compute Todini index, equations from :cite:p:`todi00`.

    The Todini index is related to the capability of a system to overcome
    failures while still meeting demands and pressures at the nodes. The
    Todini index defines resilience at a specific time as a measure of surplus
    power at each node and measures relative energy redundancy.

    Parameters
    ----------
    head : pandas DataFrame
        A pandas DataFrame containing node head 
        (index = times, columns = node names).
        
    pressure : pandas DataFrame
        A pandas DataFrame containing node pressure 
        (index = times, columns = node names).
        
    demand : pandas DataFrame
        A pandas DataFrame containing node demand 
        (index = times, columns = node names).
        
    flowrate : pandas DataFrame
        A pandas DataFrame containing pump flowrates 
        (index = times, columns = pump names).

    wn : wntr WaterNetworkModel
        Water network model.  The water network model is needed to 
        find the start and end node to each pump.
        
    Pstar : float
        Pressure threshold.

    Returns
    -------
    A pandas Series that contains a time-series of Todini indexes
    """

    Pout = demand.loc[:,wn.junction_name_list]*head.loc[:,wn.junction_name_list]
    elevation = head.loc[:,wn.junction_name_list]-pressure.loc[:,wn.junction_name_list]
    Pexp = demand.loc[:,wn.junction_name_list]*(Pstar+elevation)

    Pin_res = -demand.loc[:,wn.reservoir_name_list]*head.loc[:,wn.reservoir_name_list]

    headloss = pd.DataFrame()
    for name, link in wn.pumps():
        start_node = link.start_node_name
        end_node = link.end_node_name
        start_head = head.loc[:,start_node] # (m)
        end_head = head.loc[:,end_node] # (m)
        headloss[name] = end_head - start_head # (m)
        
    Pin_pump = flowrate.loc[:,wn.pump_name_list]*headloss.abs()

    todini = (Pout.sum(axis=1) - Pexp.sum(axis=1))/  \
        (Pin_res.sum(axis=1) + Pin_pump.sum(axis=1) - Pexp.sum(axis=1))
    
    return todini

def modified_resilience_index(pressure, elevation, Pstar, demand=None, per_junction=True):
    """
    Compute the modified resilience index, equations from :cite:p:`jasr08`.

    The modified resilience index is the total surplus power available at 
    demand junctions as a percentage of the total minimum required power at 
    demand junctions. The metric can be computed as a timeseries for each 
    junction or as a system average timeseries.

    Parameters
    ----------
    pressure : pandas DataFrame
        A pandas DataFrame containing junction pressure 
        (index = times, columns = junction names).
        
    elevation : pandas Series
        Junction elevation (which can be obtained using `wn.query_node_attribute('elevation')`)
        (index = junction names)
        
    Pstar : float
        Pressure threshold.
    
    demand : pandas DataFrame
        A pandas DataFrame containing junction demand (only needed if per_junction=False)
        (index = times, columns = junction names).

    per_junction : bool (optional)
        If True, compute the modified resilience index per junction.
        If False, compute the modified resilience index over all junctions.
        
    Returns
    -------
    pandas Series or DataFrame
        Modified resilience index time-series. If per_junction=True, columns=junction names.
    """
    assert isinstance(pressure, pd.DataFrame), "pressure must be a pandas DataFrame"
    assert isinstance(elevation, pd.Series), "elevation must be a pandas Series"
    assert sorted(pressure.columns) == sorted(elevation.index), "The columns in pressure must be the same as the index in elevation"
    assert isinstance(Pstar, (float, int)), "Pstar must be a float"
    assert isinstance(per_junction, bool), "per_junction must be a Boolean"
    if per_junction == False:
        assert isinstance(demand, pd.DataFrame), "demand must be a pandas DataFrame when per_junction=False"
        assert sorted(pressure.columns) == sorted(demand.columns), "The columns in pressure must be the same as the columns in demand"
    
    if per_junction:
        Pout = (pressure + elevation)
        Pexp =(Pstar + elevation)
        mri = (Pout - Pexp)/Pexp
    else:
        
        Pout = demand*(pressure + elevation)
        Pexp = demand*(Pstar + elevation)
        mri = (Pout.sum(axis=1) - Pexp.sum(axis=1))/Pexp.sum(axis=1)
    
    return mri

def tank_capacity(pressure, wn):
    """
    Compute tank capacity, the ratio of water volume stored in tanks to the 
    maximum volume of water that can be stored.

    Parameters
    ----------
    pressure : pandas DataFrame
        A pandas DataFrame containing tank water level (pressure) 
        (index = times, columns = tank names).
        
    wn : wntr WaterNetworkModel
        Water network model.  The water network model is needed to 
        get the tank object to compute current and max volume.
        
    Returns
    -------
    pandas DataFrame
        Tank capacity (index = times, columns = tank names)
    """
    
    assert isinstance(pressure, pd.DataFrame), "pressure must be a pandas DataFrame"
    assert isinstance(wn, wntr.network.WaterNetworkModel), "wn must be a wntr WaterNetworkModel"

    tank_capacity = pd.DataFrame(index=pressure.index, columns=pressure.columns)
    
    for name in wn.tank_name_list:
        tank = wn.get_node(name)
        max_volume = tank.get_volume(tank.max_level)
        tank_volume = tank.get_volume(pressure[name])
        tank_capacity[name] = tank_volume/max_volume
    
    return tank_capacity
    
def entropy(G, sources=None, sinks=None):
    """
    Compute entropy, equations from :cite:p:`awgb90`.

    Entropy is a measure of uncertainty in a random variable.
    In a water distribution network model, the random variable is
    flow in the pipes and entropy can be used to measure alternate flow paths
    when a network component fails.  A network that carries maximum entropy
    flow is considered reliable with multiple alternate paths.

    Parameters
    ----------
    G : NetworkX or WNTR graph
        Entropy is computed using a directed graph based on pipe flow direction.
        The 'weight' of each link is equal to the flow rate.

    sources : list of strings, optional (default = all reservoirs)
        List of node names to use as sources.

    sinks : list of strings, optional (default = all nodes)
        List of node names to use as sinks.

    Returns
    -------
    A tuple which includes:
        - A pandas Series that contains entropy for each node
        - System entropy (float)
    """

    if G.is_directed() == False:
        return

    if sources is None:
        sources = [key for key,value in nx.get_node_attributes(G,'type').items() if value == 'Reservoir' ]

    if sinks is None:
        sinks = G.nodes()

    S = {}
    Q = {}
    for nodej in sinks:
        if nodej in sources:
            S[nodej] = 0 # nodej is the source
            continue

        sp = [] # simple path
        if G.nodes[nodej]['type']  == 'Junction':
            for source in sources:
                if nx.has_path(G, source, nodej):
                    simple_paths = nx.all_simple_paths(G,source,target=nodej)
                    sp = sp + ([p for p in simple_paths])
                    # all_simple_paths was modified to check 'has_path' in the
                    # loop, but this is still slow for large networks
                    # what if the network was skeletonized based on series pipes
                    # that have the same flow direction?
                    # what about duplicating paths that have pipes in series?
                #print j, nodeid, len(sp)

        if len(sp) == 0:
            S[nodej] = np.nan # nodej is not connected to any sources
            continue

        # "dtype=object" is needed to create an array from a list of lists with differnet lengths
        sp = np.array(sp, dtype=object)

        # Uj = set of nodes on the upstream ends of links incident on node j
        Uj = G.predecessors(nodej)
        # qij = flow in link from node i to node j
        qij = []
        # aij = number of equivalnet independent paths through the link from node i to node j
        aij = []
        for nodei in Uj:
            mask = np.array([nodei in path for path in sp])
            # NDij = number of paths through the link from node i to node j
            NDij = sum(mask)
            if NDij == 0:
                continue
            temp = sp[mask]
            # MDij = links in the NDij path
            MDij = [(t[idx],t[idx+1]) for t in temp for idx in range(len(t)-1)]

            flow = 0
            for link in G[nodei][nodej].keys():
                flow = flow + G[nodei][nodej][link]['weight']
            qij.append(flow)

            # dk = degree of link k in MDij
            dk = Counter()
            for elem in MDij:
                # divide by the numnber of links between two nodes
                dk[elem] += 1/len(G[elem[0]][elem[1]].keys())
            V = np.array(list(dk.values()))
            aij.append(NDij*(1-float(sum(V - 1))/sum(V)))

        Q[nodej] = sum(qij) # Total flow into node j

        # Equation 7
        S[nodej] = 0
        for idx in range(len(qij)):
            if Q[nodej] != 0 and qij[idx]/Q[nodej] > 0:
                S[nodej] = S[nodej] - \
                    qij[idx]/Q[nodej]*math.log(qij[idx]/Q[nodej]) + \
                    qij[idx]/Q[nodej]*math.log(aij[idx])

    Q0 = sum(nx.get_edge_attributes(G, 'weight').values())

    # Equation 3
    S_ave = 0
    for nodej in sinks:
        if not np.isnan(S[nodej]):
            if nodej not in sources:
                if Q[nodej]/Q0 > 0:
                    S_ave = S_ave + \
                        (Q[nodej]*S[nodej])/Q0 - \
                        Q[nodej]/Q0*math.log(Q[nodej]/Q0)
                        
    S = pd.Series(S) # convert S to a series
    
    return [S, S_ave]



"""
The wntr.gis.network module contains methods to convert between water network 
models and GIS formatted data
"""
from __future__ import annotations
import pandas as pd
import numpy as np

from pathlib import Path
from typing import Literal

try:
    from shapely.geometry import LineString, Point
    has_shapely = True
except ModuleNotFoundError:
    has_shapely = False

try:
    import geopandas as gpd
    has_geopandas = True
except ModuleNotFoundError:
    has_geopandas = False

import wntr.network.elements

class WaterNetworkGIS:
    """
    Water network GIS class 
    
    Contains methods to create GeoDataFrames from WaterNetworkModel and 
    create WaterNetworkModel from GeoDataFrames.
    
    Parameters
    ----------
    gis_data : dict, optional
        Dictionary of GeoDataFrames containing data to populate an instance 
        of WaterNetworkGIS.  Valid dictionary keys are 'junction', 'tanks',
        'reservoirs', 'pipes', 'pumps', and 'valves'
    
    Raises
    ------
    ModuleNotFoundError
        if missing either shapely or geopandas
    """
    
    def __init__(self, gis_data=None) -> None:
        
        if not has_shapely or not has_geopandas:
            raise ModuleNotFoundError('shapley and geopandas are required')
        
        self.junctions = gpd.GeoDataFrame()
        self.tanks = gpd.GeoDataFrame()
        self.reservoirs = gpd.GeoDataFrame()
        self.pipes = gpd.GeoDataFrame()
        self.pumps = gpd.GeoDataFrame()
        self.valves = gpd.GeoDataFrame()
        
        if isinstance(gis_data, dict):
            if 'junctions' in gis_data.keys():
                assert isinstance(gis_data['junctions'], gpd.GeoDataFrame)
                self.junctions = gis_data['junctions']
                    
            if 'tanks' in gis_data.keys():
                assert isinstance(gis_data['tanks'], gpd.GeoDataFrame)
                self.tanks = gis_data['tanks']
                    
            if 'reservoirs' in gis_data.keys():
                assert isinstance(gis_data['reservoirs'], gpd.GeoDataFrame)
                self.reservoirs = gis_data['reservoirs']
                    
            if 'pipes' in gis_data.keys():
                assert isinstance(gis_data['pipes'], gpd.GeoDataFrame)
                self.pipes = gis_data['pipes']
                    
            if 'pumps' in gis_data.keys():
                assert isinstance(gis_data['pumps'], gpd.GeoDataFrame)
                self.pumps = gis_data['pumps']
                    
            if 'valves' in gis_data.keys():
                assert isinstance(gis_data['valves'], gpd.GeoDataFrame)
                self.valves = gis_data['valves']

    def _create_gis(self, wn, crs: str = None, pumps_as_points: bool = False, 
                   valves_as_points: bool = False,) -> None:
        """
        Create GIS data from a water network model.
        
        This method is used by wntr.network.io.to_gis
        
        Note: patterns, curves, rules, controls, sources, and options are not 
        saved to the GIS data

        Parameters
        ----------
        wn : WaterNetworkModel
            Water network model
        crs : str, optional
            Coordinate reference system, by default None
        pumps_as_points : bool, optional
            Represent pumps as points (True) or lines (False), by default False
        valves_as_points : bool, optional
            Represent valves as points (True) or lines (False), by default False
        """
        
        def _extract_geodataframe(df, crs=None, valid_base_names=None, 
                                  links_as_points=False):
            if valid_base_names is None:
                valid_base_names = []
            
            # Drop any column with all NaN, this removes excess attributes
            # Valid base attributes that have all None values are added back 
            # at the end of this routine
            df = df.loc[:, ~df.isna().all()]
            
            # Define geom and drop node_type/link_type
            if df.shape[0] > 0:
                if 'node_type' in df.columns:
                    geom = [Point((x,y)) for x,y in df['coordinates']]
                    del df['node_type']
                    
                    # do not carry over leak attributes to dataframe. 
                    del df['leak']
                    del df['leak_area']
                    del df['leak_discharge_coeff']
                elif 'link_type' in df.columns:
                    geom = []
                    for link_name in df['name']:
                        link = wn.get_link(link_name)
                        if links_as_points: #Point
                            geom.append(Point(link.start_node.coordinates))
                        else: # LineString
                            ls = list()
                            ls.append(link.start_node.coordinates)
                            for v in link.vertices:
                                ls.append(v)
                            ls.append(link.end_node.coordinates)
                            geom.append(LineString(ls))
                    del df['link_type']
                    
                # Drop column if not a str, float, int, or bool (or np.bool_)
                # This drops columns like coordinates, vertices
                # This could be extended to keep additional data type (list, 
                # tuple, network elements like Patterns, Curves)
                drop_cols = []
                for col in df.columns:
                    # Added np.bool_ to the following check
                    # Returned by df.to_dict('records') for some network models
                    if not isinstance(df.iloc[0][col], (str, float, int, bool, np.bool_)):
                        drop_cols.append(col) 
                df = df.drop(columns=drop_cols)
                
                # Add back in valid base attributes that had all None values
                cols = list(set(valid_base_names) - set(df.columns))
                cols.sort()
                if len(cols) > 0:
                    df[cols] = None
                    
                # Set index
                if len(df) > 0:
                    df.set_index('name', inplace=True)
                
                df = gpd.GeoDataFrame(df, crs=crs, geometry=geom)
            else:
                df = gpd.GeoDataFrame()

            return df
        
        # Convert the WaterNetworkModel to a dictionary
        wn_dict = wn.to_dict()
        # Create dataframes for node and link attributes
        df_nodes = pd.DataFrame(wn_dict['nodes'])
        df_links = pd.DataFrame(wn_dict['links'])
        
        valid_base_names = self._valid_names(complete_list=False, truncate_names=None)
        
        # Junctions
        df = df_nodes[df_nodes['node_type'] == 'Junction']
        self.junctions = _extract_geodataframe(df, crs, valid_base_names['junctions'])
        
        # Tanks
        df = df_nodes[df_nodes['node_type'] == 'Tank']
        self.tanks = _extract_geodataframe(df, crs, valid_base_names['tanks'])
            
        # Reservoirs
        df = df_nodes[df_nodes['node_type'] == 'Reservoir']
        self.reservoirs = _extract_geodataframe(df, crs, valid_base_names['reservoirs'])
            
        # Pipes
        df = df_links[df_links['link_type'] == 'Pipe']
        self.pipes = _extract_geodataframe(df, crs, valid_base_names['pipes'], False)
            
        # Pumps
        df = df_links[df_links['link_type'] == 'Pump']
        self.pumps = _extract_geodataframe(df, crs, valid_base_names['pumps'], pumps_as_points)
            
        # Valves
        df = df_links[df_links['link_type'] == 'Valve']
        self.valves = _extract_geodataframe(df, crs, valid_base_names['valves'], valves_as_points) 
        
    def _create_wn(self, append=None):
        """
        Create or append a WaterNetworkModel from GeoDataFrames
        
        This method is used by wntr.network.io.from_gis

        Parameters
        ----------
        append : WaterNetworkModel or None, optional
            Existing WaterNetworkModel to append.  If None, a new WaterNetworkModel 
            is created.
        """
        # Convert the WaterNetworkGIS to a dictionary
        wn_dict = {}
        wn_dict['nodes'] = []
        wn_dict['links'] = []

        # Modifications to create a WaterNetworkModel from a dict
        # Reset index
        # Create coordinates/vertices from geometry
        # Add node_type/link_type 
        for node_type, element in [('Junction', self.junctions), 
                                   ('Tank', self.tanks), 
                                   ('Reservoir', self.reservoirs)]:
            if element.shape[0] > 0:
                assert (element['geometry'].geom_type).isin(['Point']).all()
                df = element.reset_index(names="name")
                df.rename(columns={'geometry':'coordinates'}, inplace=True)
                df['coordinates'] = [[x,y] for x,y in zip(df['coordinates'].x, 
                                                          df['coordinates'].y)]
                df['node_type'] = node_type
                wn_dict['nodes'].extend(df.to_dict('records'))

        for link_type, element in [('Pipe', self.pipes), 
                                   ('Pump', self.pumps), 
                                   ('Valve', self.valves)]:
            if element.shape[0] > 0:
                assert 'start_node_name' in element.columns
                assert 'end_node_name' in element.columns
                df = element.reset_index(names="name")
                df['vertices'] = df.apply(lambda row: list(row.geometry.coords)[1:-1], axis=1)
                df.drop(columns=['geometry'], inplace=True)
                df['link_type'] = link_type
                wn_dict['links'].extend(df.to_dict('records'))
        
        # Create WaterNetworkModel from dictionary
        from wntr.network import from_dict
        wn = from_dict(wn_dict, append)
        
        return wn

    def to_crs(self, crs):
        """
        Transform CRS of the junctions, tanks, reservoirs, pipes, pumps,
        and valves GeoDataFrames.

        Calls geopandas.GeoDataFrame.to_crs on each GeoDataFrame.

        Parameters
        ----------
        crs : str
            Coordinate reference system
        """
        for data in [self.junctions, self.tanks, self.reservoirs,
                     self.pipes, self.pumps, self.valves]:
            if 'geometry' in data.columns:
                data = data.to_crs(crs, inplace=True)

    def set_crs(self, crs, allow_override=False):
        """
        Set CRS of the junctions, tanks, reservoirs, pipes, pumps,
        and valves GeoDataFrames.

        Calls geopandas.GeoDataFrame.set_crs on each GeoDataFrame.

        Parameters
        ----------
        crs : str
            Coordinate reference system
        allow_override : bool (optional)
            Allow override of existing coordinate reference system
        """

        for data in [self.junctions, self.tanks, self.reservoirs,
                     self.pipes, self.pumps, self.valves]:
            if 'geometry' in data.columns:
                data = data.set_crs(crs, inplace=True,
                                    allow_override=allow_override)

    def add_node_attributes(self, values, name):
        """
        Add attribute to junctions, tanks, or reservoirs GeoDataFrames

        Parameters
        ----------
        values : dict or Series or row of a DataFrame
            Attribute values
        name : str
            Attribute name
        """
        for node_name, value in values.items():
            if node_name in self.junctions.index:
                if name not in self.junctions.columns:
                    self.junctions[name] = np.nan
                self.junctions.loc[node_name, name] = value
            elif node_name in self.tanks.index:
                if name not in self.tanks.columns:
                    self.tanks[name] = np.nan
                self.tanks.loc[node_name, name] = value
            elif node_name in self.reservoirs.index:
                if name not in self.reservoirs.columns:
                    self.reservoirs[name] = np.nan
                self.reservoirs.loc[node_name, name] = value

    def add_link_attributes(self, values, name):
        """
        Add attribute to pipes, pumps, or valves GeoDataFrames

        Parameters
        ----------
        values : dict or Series or row of a DataFrame
            Attribute values
        name : str
            Attribute name
        """
        for link_name, value in values.items():
            if link_name in self.pipes.index:
                if name not in self.pipes.columns:
                    self.pipes[name] = np.nan
                self.pipes.loc[link_name, name] = value
            elif link_name in self.valves.index:
                if name not in self.valves.columns:
                    self.valves[name] = np.nan
                self.valves.loc[link_name, name] = value
            elif link_name in self.pumps.index:
                if name not in self.pumps.columns:
                    self.pumps[name] = np.nan
                self.pumps.loc[link_name, name] = value
    
    def _read(self, files, index_col='name'):
        
        for layer,file in files.items():
            p = Path(file)
            if p.suffix:
                continue
            files[layer] = str(p / (p.name + ".shp"))


        if 'junctions' in files.keys():
            data = gpd.read_file(files['junctions']).set_index(index_col)
            self.junctions = pd.concat([self.junctions, data])
        if 'tanks' in files.keys():
            data = gpd.read_file(files['tanks']).set_index(index_col)
            self.tanks = pd.concat([self.tanks, data])
        if 'reservoirs' in files.keys():
            data = gpd.read_file(files['reservoirs']).set_index(index_col)
            self.reservoirs = pd.concat([self.reservoirs, data])
        if 'pipes' in files.keys():
            data = gpd.read_file(files['pipes']).set_index(index_col)
            self.pipes = pd.concat([self.pipes, data])
        if 'pumps' in files.keys():
            data = gpd.read_file(files['pumps']).set_index(index_col)
            self.pumps = pd.concat([self.pumps, data])
        if 'valves' in files.keys():
            data = gpd.read_file(files['valves']).set_index(index_col)
            self.valves = pd.concat([self.valves, data])

    def read_geojson(self, files, index_col='name'):
        """
        Append information from GeoJSON files to a WaterNetworkGIS object

        Parameters
        ----------
        files : dictionary
            Dictionary of GeoJSON filenames, where the keys are in the set 
            ('junction', 'tanks', 'reservoirs', 'pipes', 'pumps', 'valves') and 
            values are the corresponding GeoJSON filename
        index_col : str, optional
            Column that contains the element name
        """
        self._read(files, index_col)

    def read_shapefile(self, files, index_col='name'):
        """
        Append information from Esri Shapefiles to a WaterNetworkGIS object

        Parameters
        ----------
        files : dictionary
            Dictionary of Shapefile directory or filenames, where the keys are
            in the set ('junction', 'tanks', 'reservoirs', 'pipes', 'pumps',
            'valves') and values are the corresponding GeoJSON filename
        index_col : str, optional
            Column that contains the element name
        """
        self._read(files, index_col)

        field_name_map = self._shapefile_field_name_map()

        self.junctions.rename(columns=field_name_map['junctions'], inplace=True)
        self.tanks.rename(columns=field_name_map['tanks'], inplace=True)
        self.reservoirs.rename(columns=field_name_map['reservoirs'], inplace=True)
        self.pipes.rename(columns=field_name_map['pipes'], inplace=True)
        self.pumps.rename(columns=field_name_map['pumps'], inplace=True)
        self.valves.rename(columns=field_name_map['valves'], inplace=True)

    def _write(self, prefix: str, driver="GeoJSON") -> None:
        """
        Write the WaterNetworkGIS object to GIS files

        One file will be created for each type of network element (junctions, 
        pipes, etc.) if those elements exists in the network
        
        Parameters
        ----------
        prefix : str
            Filename prefix, will have the element type (junctions, 
            pipes, etc.) appended
        driver : str, optional
            GeoPandas driver. Use "GeoJSON" for GeoJSON files, use :code:`None` 
            for Esri Shapefile folders, by default "GeoJSON"

        """
        
        if len(self.junctions) > 0:
            filename = self._get_filename(driver,prefix,'junctions')
            self.junctions.to_file(filename, driver=driver)
                
        if len(self.tanks) > 0:
            filename = self._get_filename(driver,prefix,'tanks')
            self.tanks.to_file(filename, driver=driver)
            
        if len(self.reservoirs) > 0:
            filename = self._get_filename(driver,prefix,'reservoirs')
            self.reservoirs.to_file(filename, driver=driver)
            
        if len(self.pipes) > 0:
            filename = self._get_filename(driver,prefix,'pipes')
            self.pipes.to_file(filename, driver=driver)
            
        if len(self.pumps) > 0:
            filename = self._get_filename(driver,prefix,'pumps')
            self.pumps.to_file(filename, driver=driver)
            
        if len(self.valves) > 0:
            filename = self._get_filename(driver,prefix,'valves')
            self.valves.to_file(filename, driver=driver)

    def _get_filename(self, driver: Literal['GeoJSON'|'shp']|None, prefix: str|Path, elements:str) -> str:
        
        if driver:
            return str(prefix) + "_" + elements + "." + driver.lower()

        else:
            prefix = Path(prefix)

            directory = Path(str(prefix) + '_' + elements)
            directory.mkdir(parents=True, exist_ok=True)

            return str( directory / ( prefix.name + '_' + elements + ".shp"))


    def write_geojson(self, prefix: str):
        """
        Write the WaterNetworkGIS object to a set of GeoJSON files, one file
        for each network element.

        Parameters
        ----------
        prefix : str
            File prefix
        """
        self._write(prefix=prefix, driver="GeoJSON")

    def write_shapefile(self, prefix: str):
        """
        Write the WaterNetworkGIS object to a set of Esri Shapefiles, one
        directory for each network element.

        Parameters
        ----------
        prefix : str
            File and directory prefix
        """
        self._write(prefix=prefix, driver=None)

    def _valid_names(self, complete_list=True, truncate_names=None):
        """
        Valid column/field names for GeoJSON or Shapefiles
        
        Note that Shapefile field names are truncated to 10 characters 
        (set truncate=10)
        
        Parameters
        ----------
        complete_list : bool
            Include a complete list of column/field names (beyond basic attributes)
        truncate_names : None or int
            Truncate column/field names to specified number of characters, 
            set truncate=10 for Shapefiles.  None indicates no truncation.
            
        Returns
        ---------
        dict : Dictionary of valid GeoJSON or Shapefile column/field names
        """
        
        valid_names = {}
          
        element_objects = {
            'junctions': wntr.network.elements.Junction,
            'tanks': wntr.network.elements.Tank,
            'reservoirs': wntr.network.elements.Reservoir,
            'pipes': wntr.network.elements.Pipe,
            'pumps': wntr.network.elements.Pump,
            'valves': wntr.network.elements.Valve}
        
        valid_names = {}
        for element, obj in element_objects.items():
            if complete_list:
                valid_names[element] = obj._base_attributes + obj._optional_attributes
            else:
                valid_names[element] = obj._base_attributes
                
        if truncate_names is not None and truncate_names > 0:
            for element, attributes in valid_names.items():
                valid_names[element] = [attribute[:truncate_names] for attribute in attributes]
        
        for key, vals in valid_names.items():
            # Remove coordinates and vertices (not used to create GeoDataFrame geometry)
            if 'coordinates' in valid_names[key]:
                valid_names[key].remove('coordinates')
            if 'vertices' in valid_names[key]:
                valid_names[key].remove('vertices')
            
            # Add geometry
            if 'geometry' not in valid_names[key]:
                valid_names[key].append('geometry')

        return valid_names
    
    def _shapefile_field_name_map(self):
        """
        Return a map (dictionary) of tuncated shapefile field names to 
        valid base WaterNetworkModel attribute names
        
        Esri Shapefiles truncate field names to 10 characters. The field name 
        map links truncated shapefile field names to complete (and ofen longer)
        WaterNetworkModel attribute names.  This assumes that the first 10 
        characters of each attribute name are unique.
        
        Returns
        -------
        field_name_map : dict
            Map (dictionary) of valid base shapefile field names to 
            WaterNetworkModel attribute names
        """
        valid_names = self._valid_names()

        field_name_map = {}
        for element, attributes in valid_names.items():
            truncated = [attribute[:10] for attribute in attributes]
            field_name_map[element] = pd.Series(dict(zip(truncated, attributes)))

        return field_name_map



"""
The wntr.graphics.layer module includes methods plot
water network data layers.
"""
import logging
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)

def plot_valve_layer(wn, valve_layer, valve_attribute=None, title=None,
               valve_size=15, valve_range=[None,None], valve_cmap=None, add_colorbar=True, 
               colorbar_label=None, include_network=True, ax=None, filename=None):
    """
    Plot valve layer
    
    Parameters
    ----------
    wn: wntr WaterNetworkModel
        A WaterNetworkModel object
    
    valve_layer: pd.Dataframe, optional
        Valve layer, defined by node and link pairs (for example, valve 0 is 
        on link A and protects node B). The valve_layer DataFrame is indexed by
        valve number, with columns named 'node' and 'link'.
    
    valve_attribute: pd.Series, optional
        Attribute for each valve
        
    title: str, optional
        Plot title 

    valve_size: int, optional
        Node size 
    
    valve_range: list, optional
        Value range used to scale colormap ([None,None] indicates autoscale)
        
    valve_cmap: matplotlib.pyplot.cm colormap or named color, optional
        Valve colormap 

    add_colorbar: bool, optional
        Add colorbar
    
    colorbar_label: str, optional
        Node colorbar label
    
    include_network: bool, options
        Include a plot of the water network 
        
    ax: matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with a single 
        axes will be used)
        
    filename : str, optional
        Filename used to save the figure
        
    Returns
    -------
    ax : matplotlib axes object  
    """

    if ax is None: # create a new figure
        plt.figure(facecolor='w', edgecolor='k')
        ax = plt.gca()
        
    # Graph, undirected
    G = wn.to_graph().to_undirected()

    # Position
    pos = nx.get_node_attributes(G,'pos')
    if len(pos) == 0:
        pos = None

    if title is not None:
        ax.set_title(title)
    
    if include_network:
        nx.draw_networkx_edges(G, pos, edge_color='grey', width=0.5, ax=ax)
        
    ax.axis('off')
    
    valve_coordinates = []
    for valve_name, (pipe_name, node_name) in valve_layer.iterrows():
        pipe = wn.get_link(pipe_name)
        if node_name == pipe.start_node_name:
            start_node = pipe.start_node
            end_node = pipe.end_node
        elif node_name == pipe.end_node_name:
            start_node = pipe.end_node
            end_node = pipe.start_node
        else:
            print("Not valid")
            continue
        x0 = start_node.coordinates[0]
        dx = end_node.coordinates[0] - x0
        y0 = start_node.coordinates[1]
        dy = end_node.coordinates[1] - y0
        valve_coordinates.append((x0 + dx * 0.1, y0 + dy * 0.1))
    
    valve_coordinates = np.array(valve_coordinates)
    
    if valve_attribute is not None: 
        sc = ax.scatter(valve_coordinates[:,0], valve_coordinates[:,1], s=valve_size, c=valve_attribute, marker='v', cmap=valve_cmap)  
    else:
        sc = ax.scatter(valve_coordinates[:,0], valve_coordinates[:,1], valve_size, 'k', 'v')   
 
    if add_colorbar:
        clb = plt.colorbar(sc, shrink=0.5, pad=0.05, ax=ax)
        clb.ax.set_title(colorbar_label, fontsize=10)
        clb.mappable.set_clim(valve_range[0],valve_range[1])
        
    if filename:
        plt.savefig(filename)
        
    return ax



"""
The wntr.metrics.economic module contains economic metrics.
"""
from wntr.network import Tank, Pipe, Pump, Valve
import numpy as np 
import pandas as pd 
import logging
import scipy

logger = logging.getLogger(__name__)

def annual_network_cost(wn, tank_cost=None, pipe_cost=None, prv_cost=None, 
                        pump_cost=None):
    """ 
    Compute annual network cost :cite:p:`sokz12`.
    
    Use the closest value from the lookup tables to compute annual cost for each 
    component in the network.
    
    Parameters
    ----------
    wn : wntr WaterNetworkModel
        Water network model. The water network model is needed to 
        define tank volume, pipe and valve diameter, and pump power conditions.
    
    tank_cost : pandas Series, optional
        Annual tank cost indexed by volume 
        (default values below, from :cite:p:`sokz12`).
    
        =============  ================================
        Volume (m3)    Annual Cost ($/yr) 
        =============  ================================
        500             14020
        1000            30640
        2000            61210
        3750            87460
        5000            122420
        10000           174930
        =============  ================================
    
    pipe_cost : pandas Series, optional
        Annual pipe cost per pipe length indexed by diameter
        (default values below, from :cite:p:`sokz12`).
    
        =============  =============  ================================
        Diameter (in)   Diameter (m)  Annual Cost ($/m/yr) 
        =============  =============  ================================
        4              0.102          8.31
        6              0.152          10.10
        8              0.203          12.10
        10             0.254          12.96
        12             0.305          15.22
        14             0.356          16.62
        16             0.406          19.41
        18             0.457          22.20
        20             0.508          24.66
        24             0.610          35.69
        28             0.711          40.08
        30             0.762          42.60
        =============  =============  ================================
        
    prv_cost : pandas Series, optional
        Annual PRV valve cost indexed by diameter 
        (default values below, from :cite:p:`sokz12`).
        
        =============  =============  ================================
        Diameter (in)   Diameter (m)  Annual Cost ($/m/yr) 
        =============  =============  ================================
        4              0.102          323
        6              0.152          529
        8              0.203          779
        10             0.254          1113
        12             0.305          1892
        14             0.356          2282
        16             0.406          4063
        18             0.457          4452
        20             0.508          4564
        24             0.610          5287
        28             0.711          6122
        30             0.762          6790
        =============  =============  ================================
    
    pump_cost : pd.Series, optional
        Annual pump cost indexed by maximum power input to pump
        (default values below, from :cite:p:`sokz12`).
        Maximum Power for a HeadPump is computed from the pump curve
        as follows:
        
        .. math:: Pmp = g*rho/eff*exp(ln(A/(B*(C+1)))/C)*(A - B*(exp(ln(A/(B*(C+1)))/C))^C)
        
        where 
        :math:`Pmp` is the maximum power (W), 
        :math:`g` is acceleration due to gravity (9.81 m/s^2), 
        :math:`rho` is the density of water (1000 kg/m^3), 
        :math:`eff` is the global efficiency (0.75 default),
        :math:`A`, :math:`B`, and :math:`C` are the pump curve coefficients.

        ==================  ================================
        Maximum power (W)   Annual Cost ($/yr) 
        ==================  ================================
        11310               2850
        22620               3225
        24880               3307
        31670               3563
        38000               3820
        45240               4133
        49760               4339
        54280               4554
        59710               4823
        ==================  ================================

    Returns
    ----------
    Annual network cost in dollars (float)
    """
    
    # Initialize network construction cost
    network_cost = 0
    
    # Set defaults
    if tank_cost is None:
        volume = [500, 1000, 2000, 3750, 5000, 10000] 
        cost =  [14020, 30640, 61210, 87460, 122420, 174930]
        tank_cost = pd.Series(cost, volume)
        
    if pipe_cost is None:
        diameter = [4, 6, 8, 10, 12, 14, 16, 18, 20, 24, 28, 30] # inch
        diameter = np.array(diameter)*0.0254 # m
        cost =  [8.31, 10.1, 12.1, 12.96, 15.22, 16.62, 19.41, 22.2, 24.66, 35.69, 40.08, 42.6]
        pipe_cost = pd.Series(cost, diameter)
        
    if prv_cost is None:
        diameter = [4, 6, 8, 10, 12, 14, 16, 18, 20, 24, 28, 30] # inch
        diameter = np.array(diameter)*0.0254 # m
        cost =  [323, 529, 779, 1113, 1892, 2282, 4063, 4452, 4564, 5287, 6122, 6790]
        prv_cost = pd.Series(cost, diameter)

    if pump_cost is None:
        Pmp = [11310, 22620, 24880, 31670, 38000, 45240, 49760, 54280, 59710]
        cost =  [2850, 3225, 3307, 3563, 3820, 4133, 4339, 4554, 4823]
        pump_cost = pd.Series(cost, Pmp)

    # Tank construction cost - not only looking at direct volume capacity but also
    #                          but a rough estimate of the support structure size
    #                          below the tank.
    for node_name, node in wn.nodes(Tank):
        if node.vol_curve is None:
            tank_construction_volume = np.pi*(node.diameter/2)**2 *node.max_level
        else:
            vcurve = np.array(node.vol_curve.points)
            tank_volume = np.interp(node.max_level,vcurve[:,0],vcurve[:,1])
            avg_area = tank_volume / (node.max_level - node.min_level)
            tank_base_volume = node.min_level * avg_area
            tank_construction_volume = tank_volume + tank_base_volume 
        # choose the entry that is closest (keep the cost structure discreet)
        idx = np.argmin([np.abs(tank_cost.index - tank_construction_volume)])
        #print(node_name, tank_cost.iloc[idx])
        network_cost = network_cost + tank_cost.iloc[idx]

    # Pipe construction cost
    for link_name, link in wn.links(Pipe):
        idx = np.argmin([np.abs(pipe_cost.index - link.diameter)])
        #print(link_name, pipe_cost.iloc[idx], link.length)
        network_cost = network_cost + pipe_cost.iloc[idx]*link.length    

    # Pump construction cost
    for link_name, link in wn.head_pumps():
        coeff = link.get_head_curve_coefficients()
        A = coeff[0]
        B = coeff[1]
        C = coeff[2]
        Pmax = 9.81*1000*np.exp(np.log(A/(B*(C+1)))/C)*(A - B*(np.exp(np.log(A/(B*(C+1)))/C))**C)
        Pmax = Pmax / wn.options.energy.global_efficiency
        idx = np.argmin([np.abs(pump_cost.index - Pmax)])
        #print(link_name, Pmax, pump_cost.iloc[idx])
        network_cost = network_cost + pump_cost.iloc[idx]

    for link_name, link in wn.power_pumps():
        Pmax = link.power
        Pmax = Pmax / wn.options.energy.global_efficiency
        idx = np.argmin([np.abs(pump_cost.index - Pmax)])
        #print(link_name, Pmax, pump_cost.iloc[idx])
        network_cost = network_cost + pump_cost.iloc[idx]
        
    # PRV valve construction cost    
    for link_name, link in wn.links(Valve):        
        if link.valve_type == 'PRV':
            idx = np.argmin([np.abs(prv_cost.index - link.diameter)])
            #print(link_name, link.diameter, prv_cost.iloc[idx])
            network_cost = network_cost + prv_cost.iloc[idx]  

    return network_cost

def annual_ghg_emissions(wn, pipe_ghg=None):
    """ 
    Compute annual greenhouse gas emissions :cite:p:`sokz12`.
    
    Use the closest value in the lookup table to compute annual GHG emissions 
    for each pipe in the network.
    
    Parameters
    ----------
    wn : wntr WaterNetworkModel
        Water network model. The water network model is needed to 
        define pipe diameter.
    
    pipe_ghg : pandas Series, optional
        Annual GHG emissions indexed by pipe diameter
        (default values below, from :cite:p:`sokz12`).
        
        =============  ================================
        Diameter (mm)  Annualized EE (kg-CO2-e/m/yr)
        =============  ================================
        102             5.90
        152             9.71
        203             13.94
        254             18.43
        305             23.16
        356             28.09
        406             33.09
        457             38.35
        508             43.76
        610             54.99
        711             66.57
        762             72.58
        =============  ================================
    
    Returns
    ----------
    Annual greenhouse gas emissions (float)
    """
    
    # Initialize network GHG emissions
    network_ghg = 0
    
    # Set defaults
    if pipe_ghg is None:
        diameter = [4, 6, 8, 10, 12, 14, 16, 18, 20, 24, 28, 30] # inches
        diameter = np.array(diameter)*0.0254 # m
        cost =  [5.9, 9.71, 13.94, 18.43, 23.16, 28.09, 33.09, 38.35, 43.76, 54.99, 66.57, 72.58]
        pipe_ghg = pd.Series(cost, diameter)
        
    # GHG emissions from pipes
    for link_name, link in wn.links(Pipe):
        idx = np.argmin([np.abs(pipe_ghg.index - link.diameter)])
        #print(link_name, link.diameter, pipe_ghg.iloc[idx],link.length)
        network_ghg = network_ghg + pipe_ghg.iloc[idx]*link.length
       
    return network_ghg    

def pump_power(flowrate, head, wn):
    """
    Compute pump power.
    
    The computation uses pump flow rate, node head (used to compute headloss at
    each pump), and pump efficiency. Pump efficiency is defined in
    ``wn.options.energy.global_efficiency``. Pump efficiency curves are currently 
    not supported.

        wn.options.energy.global_efficiency = 75 # This means 75% or 0.75

    Parameters
    ----------
    flowrate : pandas DataFrame
        A pandas DataFrame containing pump flowrates 
        (index = times, columns = pump names).
    
    head : pandas DataFrame
        A pandas DataFrame containing node head 
        (index = times, columns = node names).
        
    wn: wntr WaterNetworkModel
        Water network model.  The water network model is needed to 
        define energy efficiency.

    Returns
    -------
    A DataFrame that contains pump power in W (index = times, columns = pump names).
    """
    
    pumps = wn.pump_name_list
    time = flowrate.index
    
    headloss = pd.DataFrame(data=None, index=time, columns=pumps)
    for pump_name, pump in wn.pumps():
        start_node = pump.start_node_name
        end_node = pump.end_node_name
        start_head = head.loc[:,start_node]
        end_head = head.loc[:,end_node]
        headloss.loc[:,pump_name] = end_head - start_head

    efficiency_dict = {}
    for pump_name, pump in wn.pumps():
        if pump.efficiency_curve_name is None:
            efficiency_dict[pump_name] = [wn.options.energy.global_efficiency/100.0 for i in time]
        else:
            raise NotImplementedError('WNTR does not support pump efficiency curves yet.')
            # TODO: WNTR does not support pump efficiency curves yet
            # curve = wn.get_curve(pump.efficiency)
            # x = [point[0] for point in curve.points]
            # y = [point[1]/100.0 for point in curve.points]
            # interp = scipy.interpolate.interp1d(x, y, kind='linear')
            # efficiency_dict[pump_name] = interp(np.array(flowrate.loc[:, pump_name]))

    efficiency = pd.DataFrame(data=efficiency_dict, index=time, columns=pumps)

    power = 1000.0 * 9.81 * headloss * flowrate / efficiency # Watts = J/s

    return power

def pump_energy(flowrate, head, wn):
    """
    Compute the pump energy over time.
    
    The computation uses pump flow rate, node head (used to compute headloss at
    each pump), and pump efficiency. Pump efficiency is defined in
    ``wn.options.energy.global_efficiency``. Pump efficiency curves are currently 
    not supported.

        wn.options.energy.global_efficiency = 75 # This means 75% or 0.75

    Parameters
    ----------
    flowrate : pandas DataFrame
        A pandas DataFrame containing pump flowrates 
        (index = times, columns = pump names).
    
    head : pandas DataFrame
        A pandas DataFrame containing node head 
        (index = times, columns = node names).
        
    wn: wntr WaterNetworkModel
        Water network model.  The water network model is needed to 
        define energy efficiency.

    Returns
    -------
    A DataFrame that contains pump energy in J (index = times, columns = pump names).
    """
    
    power = pump_power(flowrate, head, wn) # Watts = J/s
    energy = power * wn.options.time.report_timestep # J = Ws
    
    return energy

def pump_cost(energy, wn):
    """
    Compute the pump cost over time. 
    
    Energy cost is defined in ``wn.options.energy.global_price``. Pump energy 
    price and price patterns are currently not supported. 

        wn.options.energy.global_price = 3.61e-8  # $/J; equal to $0.13/kW-h
        
    Parameters
    ----------
    energy : pandas DataFrame
        A pandas DataFrame containing pump energy (J), computed from ``wntr.metrics.pump_energy`` 
        (index = times, columns = pump names).
        
    wn: wntr WaterNetworkModel
        Water network model.  The water network model is needed to 
        define pump enery prices and patterns.
        
    Returns
    -----------
    A DataFrame that contains pump cost in $ (index = times, columns = pump names).
    
    """
    time = energy.index
    pumps = wn.pump_name_list
    
    # TODO: Need to get this unit tested and not just functionally tested
    if wn.options.energy.demand_charge is not None and wn.options.energy.demand_charge != 0:
        # Additional energy charge per maximum kilowatt usage
        raise ValueError('WNTR does not support demand charge yet.')
        
    price_dict = {}
    for pump_name, pump in wn.pumps():
        if pump.energy_price is None and pump.energy_pattern is None:
            if wn.options.energy.global_pattern is None:
                price_dict[pump_name] = [wn.options.energy.global_price for i in time]
            else:
                raise NotImplementedError('WNTR does not support price patterns yet.')
        elif pump.energy_pattern is None:
            if wn.options.energy.global_pattern is None:
                price_dict[pump_name] = [pump.energy_price for i in time]
            else:
                raise NotImplementedError('WNTR does not support price patterns yet.')
        else:
            raise NotImplementedError('WNTR does not support price patterns yet.')
            
    price = pd.DataFrame(data=price_dict, index=time, columns=pumps)
    
    pump_cost = energy * price
    
    return pump_cost
    



import sys
import unittest
import pytest
from os.path import abspath, dirname, join
import threading
import time
import numpy as np

import pandas

pandas.set_option("display.max_rows", 10000)

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


def compare_results(wntr_res, epa_res, abs_threshold, rel_threshold):
    abs_diff = abs(wntr_res - epa_res)
    rel_diff = abs_diff / abs(epa_res)
    abs_diffa = abs_diff.mean()
    rel_diffa = rel_diff.mean()
    abs_diffb = abs_diff.mean(axis=1)
    rel_diffb = rel_diff.mean(axis=1)
    diffa = (abs_diffa < abs_threshold) | (rel_diffa < rel_threshold)
    diffb = (abs_diffb < abs_threshold) | (rel_diffb < rel_threshold)
    return diffa.all() and diffb.all()


class TestPerformance(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

    @classmethod
    def tearDownClass(self):
        pass

    def test_Anytown_multipointcurves_performance(self):
        head_diff_abs_threshold = 1e-3
        demand_diff_abs_threshold = 1e-5
        flow_diff_abs_threshold = 1e-5
        rel_threshold = 1e-3

        inp_file = join(test_datadir, "Anytown_multipointcurves.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        wn2 = self.wntr.network.WaterNetworkModel(inp_file)

        # Apply a curve that is very fine along H = a - b * Q ** c and
        # verify the answers are nearly identical under these conditions.
        A = 1.313e3
        B = 8.705e-6
        C = 1.796e0
        new_curve_name = "perfect_curve"
        cdata = [(10 * q, A - B * (10 * q) ** C) for q in range(1200)]
        wn.add_curve(new_curve_name, "HEAD", cdata)
        wn2.add_curve(new_curve_name, "HEAD", cdata)
        for pump_name, pump in wn.pumps():
            pump.pump_curve_name = new_curve_name
        for pump_name, pump in wn2.pumps():
            pump.pump_curve_name = new_curve_name
        
        epa_sim = self.wntr.sim.EpanetSimulator(wn)
        epa_res = epa_sim.run_sim()

        sim = self.wntr.sim.WNTRSimulator(wn2)
        results = sim.run_sim()

        self.assertTrue(
            compare_results(
                results.node["head"],
                epa_res.node["head"],
                head_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.node["demand"].loc[:, wn.tank_name_list],
                epa_res.node["demand"].loc[:, wn.tank_name_list],
                demand_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.link["flowrate"],
                epa_res.link["flowrate"],
                flow_diff_abs_threshold,
                rel_threshold,
            )
        )

    def test_Net1_float64(self):
        """Only needs to test that runs successfully with float64 time options"""
        inp_file = join(ex_datadir, "Net1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        
        wn.options.time.duration = np.float64(wn.options.time.duration)
        wn.options.time.hydraulic_timestep = np.float64(wn.options.time.hydraulic_timestep)
        wn.options.time.quality_timestep = np.float64(wn.options.time.quality_timestep)
        wn.options.time.rule_timestep = np.float64(wn.options.time.rule_timestep)
        wn.options.time.pattern_timestep = np.float64(wn.options.time.pattern_timestep)
        wn.options.time.pattern_start = np.float64(wn.options.time.pattern_start)
        wn.options.time.report_timestep = np.float64(wn.options.time.report_timestep)
        wn.options.time.report_start = np.float64(wn.options.time.report_start)
        wn.options.time.start_clocktime = np.float64(wn.options.time.start_clocktime)
        
        epa_sim = self.wntr.sim.EpanetSimulator(wn)
        epa_res = epa_sim.run_sim()

        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()
    
    def test_Net1_charset(self):
        """Only needs to test that runs successfully with latin-1 character set."""
        inp_file = join(test_datadir, "latin1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)

        epa_sim = self.wntr.sim.EpanetSimulator(wn)
        epa_res = epa_sim.run_sim()

        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

    def test_Net1_performance(self):
        head_diff_abs_threshold = 1e-3
        demand_diff_abs_threshold = 1e-5
        flow_diff_abs_threshold = 1e-5
        rel_threshold = 1e-3

        inp_file = join(ex_datadir, "Net1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)

        epa_sim = self.wntr.sim.EpanetSimulator(wn)
        epa_res = epa_sim.run_sim()

        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        self.assertTrue(
            compare_results(
                results.node["head"],
                epa_res.node["head"],
                head_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.node["demand"].loc[:, wn.tank_name_list],
                epa_res.node["demand"].loc[:, wn.tank_name_list],
                demand_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.link["flowrate"],
                epa_res.link["flowrate"],
                flow_diff_abs_threshold,
                rel_threshold,
            )
        )

    def test_Net3_performance(self):
        head_diff_abs_threshold = 1e-3
        demand_diff_abs_threshold = 1e-5
        flow_diff_abs_threshold = 1e-5
        rel_threshold = 1e-3

        inp_file = join(ex_datadir, "Net3.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)

        epa_sim = self.wntr.sim.EpanetSimulator(wn)
        epa_res = epa_sim.run_sim()

        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        self.assertTrue(
            compare_results(
                results.node["head"],
                epa_res.node["head"],
                head_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.node["demand"].loc[:, wn.tank_name_list],
                epa_res.node["demand"].loc[:, wn.tank_name_list],
                demand_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.link["flowrate"],
                epa_res.link["flowrate"],
                flow_diff_abs_threshold,
                rel_threshold,
            )
        )

    def test_Net3_performance_PDA(self):
        head_diff_abs_threshold = 1e-3
        demand_diff_abs_threshold = 1e-5
        flow_diff_abs_threshold = 1e-5
        rel_threshold = 1e-3

        inp_file = join(ex_datadir, "Net3.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        wn.options.hydraulic.demand_model = "PDA"

        epa_sim = self.wntr.sim.EpanetSimulator(wn)
        epa_res = epa_sim.run_sim(version=2.2)

        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        self.assertTrue(
            compare_results(
                results.node["head"],
                epa_res.node["head"],
                head_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.node["demand"].loc[:, wn.tank_name_list],
                epa_res.node["demand"].loc[:, wn.tank_name_list],
                demand_diff_abs_threshold,
                rel_threshold,
            )
        )
        self.assertTrue(
            compare_results(
                results.link["flowrate"],
                epa_res.link["flowrate"],
                flow_diff_abs_threshold,
                rel_threshold,
            )
        )
    
    @pytest.mark.time_consuming
    def test_Net6_thread_performance(self):
        """
        Test thread-safe performance of simulators
        """
        def run_epanet(wn, name):
            sim = self.wntr.sim.EpanetSimulator(wn)
            sim.run_sim(name, version=2.2)

        def run_wntr(wn, name):
            sim = self.wntr.sim.WNTRSimulator(wn)
            sim.run_sim()

        inp_file = join(ex_datadir, "Net6.inp")
        wn1 = self.wntr.network.WaterNetworkModel(inp_file)
        wn1.options.time.duration = 24 * 3600
        wn1.options.time.hydraulic_timestep = 3600
        wn1.options.time.report_timestep = 3600
        wn1.remove_control(
            "control 72"
        )  # this control never gets activated in epanet because it uses a threshold equal to the tank max level
        wn2 = self.wntr.network.WaterNetworkModel(inp_file)
        wn2.options.time.duration = 24 * 3600
        wn2.options.time.hydraulic_timestep = 3600
        wn2.options.time.report_timestep = 3600
        wn2.remove_control(
            "control 72"
        )  # this control never gets activated in epanet because it uses a threshold equal to the tank max level

        start_time = time.time()
        run_epanet(wn1, 'temp1')
        run_epanet(wn2, 'temp2')
        seq_time = time.time()-start_time

        t1 = threading.Thread(target=run_epanet, args=(wn1, 'temp1'))
        t2 = threading.Thread(target=run_epanet, args=(wn2, 'temp2'))
        start_time = time.time()
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        thr_time = time.time()-start_time
        self.assertGreaterEqual(seq_time - thr_time, -1, 'EPANET threading took 1s longer than sequential')

        start_time = time.time()
        run_wntr(wn1, 'temp1')
        run_wntr(wn2, 'temp2')
        seq_time = time.time()-start_time

        t1 = threading.Thread(target=run_wntr, args=(wn1, 'temp1'))
        t2 = threading.Thread(target=run_wntr, args=(wn2, 'temp2'))
        start_time = time.time()
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        thr_time = time.time()-start_time
        self.assertGreaterEqual(seq_time - thr_time, -1, 'WNTR threading took 1s longer than sequential')
        
    @pytest.mark.time_consuming
    def test_Net6_mod_performance(self):
        head_diff_abs_threshold = 1e-3
        demand_diff_abs_threshold = 1e-5
        flow_diff_abs_threshold = 1e-5
        rel_threshold = 1e-3

        inp_file = join(ex_datadir, "Net6.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.duration = 24 * 3600
        wn.options.time.hydraulic_timestep = 3600
        wn.options.time.report_timestep = 3600
        wn.remove_control(
            "control 72"
        )  # this control never gets activated in epanet because it uses a threshold equal to the tank max level

        epa_sim = self.wntr.sim.EpanetSimulator(wn)
        epa_res = epa_sim.run_sim()

        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        self.assertTrue(
            compare_results(
                results.node["head"],
                epa_res.node["head"],
                head_diff_abs_threshold,
                rel_threshold,
            )
        )
        # self.assertTrue(compare_results(results.node['demand'].loc[:, wn.tank_name_list], epa_res.node['demand'].loc[:, wn.tank_name_list], demand_diff_abs_threshold, rel_threshold))
        # self.assertTrue(compare_results(results.link['flowrate'], epa_res.link['flowrate'], flow_diff_abs_threshold, rel_threshold))


if __name__ == "__main__":
    unittest.main()



import math
import unittest
from os.path import abspath, dirname, join

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestPDD(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

    @classmethod
    def tearDownClass(self):
        pass

    def test_pdd_with_wntr(self):
        inp_file = join(test_datadir, "simulator.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        res1 = wn.get_node("reservoir1")
        res1.head_timeseries.base_value = 10.0
        p1 = wn.get_link("pipe1")
        p1.length = 0.0
        p2 = wn.get_link("pipe2")
        p2.length = 0.0

        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0

        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        for t in results.time:
            self.assertEqual(
                results.node["demand"].at[t, "junction2"],
                150.0 / 3600.0 * math.sqrt((10.0 - 0.0) / (15.0 - 0.0)),
            )


if __name__ == "__main__":
    unittest.main()



"""Contant values used by WNTRSimulator."""

import logging
from wntr.utils.polynomial_interpolation import cubic_spline

logger = logging.getLogger(__name__)


def hazen_williams_constants(m):
    m.hw_k = 10.666829500036352
    m.hw_exp = 1.852
    m.hw_minor_exp = 2

    m.hw_q1 = 0.0002
    m.hw_q2 = 0.0004
    m.hw_m = 0.001

    x1 = m.hw_q1
    x2 = m.hw_q2
    f1 = m.hw_m * m.hw_q1
    f2 = m.hw_q2 ** m.hw_exp
    df1 = m.hw_m
    df2 = m.hw_exp * m.hw_q2 ** (m.hw_exp - 1)
    a, b, c, d = cubic_spline(x1, x2, f1, f2, df1, df2)
    m.hw_a = a
    m.hw_b = b
    m.hw_c = c
    m.hw_d = d


def darcy_weisbach_constants(m):
    m.dw_k = 0.0826


def pdd_constants(m):
    m.pdd_smoothing_delta = 0.05
    m.pdd_slope = 1e-11


def head_pump_constants(m):
    m.pump_q1 = 0.0
    m.pump_q2 = 1e-8
    m.pump_slope = -1e-11


def leak_constants(m):
    m.leak_delta = 1e-4
    m.leak_slope = 1e-11



import matplotlib.pylab as plt
import pandas as pd
import numpy as np
import scipy.stats
import json
from os.path import abspath, dirname, join

from wntr.network.elements import Pattern
from wntr.network.options import TimeOptions

libdir = dirname(abspath(str(__file__)))
NoneType = type(None)

class DemandPatternLibrary(object):
    """
    Demand pattern library class.
    
    Parameters
    -------------------
    filename_or_data: str (optional)
        Filename or JSON data containing a demand pattern library.
        If None, the default DemandPatternLibrary.json file is loaded.
    """
    
    def __init__(self, filename_or_data=None):
        
        self.library = {}
        if isinstance(filename_or_data, NoneType):
            filename = join(libdir, 'DemandPatternLibrary.json')
            with open(filename, "r") as fin:
                data = json.load(fin)
        elif isinstance(filename_or_data, str):
            filename = filename_or_data
            with open(filename, "r") as fin:
                data = json.load(fin)
        elif isinstance(filename_or_data, list):
            data = filename_or_data
        
        for entry in data:
            name = entry['name']
            self.add_pattern(name, entry)
    
    @property
    def pattern_name_list(self):
        """
        Return a list of demand pattern entry names
        
        Returns
        -------
        list of strings
        """
        
        return list(self.library.keys())
    
    def get_pattern(self, name):
        """
        Return a pattern entry from the demand pattern library
        
        Parameters
        ----------
        name : str
            Pattern name
            
        Returns
        -------
        dictionary
        """
        
        assert isinstance(name, str)
        
        return self.library[name]
    
    def remove_pattern(self, name):
        """
        Remove a pattern from the demand pattern library
        
        Parameters
        ----------
        name : str
            Pattern name
        """
        
        assert isinstance(name, str)
        assert name in set(self.pattern_name_list)
        
        del self.library[name]
    
    def copy_pattern(self, name, new_name):
        """
        Add a copy of an existing pattern to the library
        
        Parameters
        ----------
        name : str
            Existing pattern name
        new_name : str
            New pattern name
        """
        
        assert isinstance(name, str)
        assert isinstance(new_name, str)
        assert name != new_name

        entry = self.get_pattern(name).copy()
        entry['name'] = new_name
        self.add_pattern(new_name, entry)

    def filter_by_category(self, category):
        """
        Return a subset of the library, filtered by category
        
        Parameters
        ----------
        category : str
            Category name
            
        Returns
        -------
        list of dictionaries
        """
        
        assert isinstance(category, (str, NoneType))

        subset = []
        for name in self.pattern_name_list:
            entry = self.get_pattern(name)
            if entry['category'] == category:
                subset.append(entry)

        return subset

    def normalize_pattern(self, name, inplace=True):
        """
        Normalize values in a pattern so the mean equals 1
        
        Parameters
        ----------
        name : str
            Pattern name
        inplace : bool
            Indicates if the pattern should be modified in place
        
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(inplace, bool)
        
        series = self.to_Series(name, duration=None)
        if series.min() < 0:
            series = series - series.min()
        series = series/series.mean()
        
        if inplace:
            self.library[name]['multipliers'] = list(series)
        
        return series
    
    def apply_noise(self, name, std, normalize=False, seed=None, inplace=True):
        """
        Apply gaussian random noise to a pattern
        
        Parameters
        ----------
        name : str
            Pattern name
        std : int or float
            Standard deviation
        normalize : bool
            Indicates if the pattern should be normalized
        seed : int
            Seed for the gaussian distribution
        inplace : bool
            Indicates if the pattern should be modified in place
        
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(std, (int, float))
        assert isinstance(normalize, bool)
        assert isinstance(seed, (int, NoneType))
        assert isinstance(inplace, bool)
        
        np.random.seed(seed)
        
        series = self.to_Series(name, duration=None)
        
        noise = np.random.normal(0, std, len(series))
        series = series + noise

        if normalize:
            series = series/series.mean()

        if inplace:
            self.library[name]['multipliers'] = list(series)
        
        return series
        
    def resample_multipliers(self, name, duration=86400, 
                             pattern_timestep=3600, start_clocktime=0,
                             wrap=True, inplace=True):
        """
        Resample multipliers, which can change if the start_clocktime, 
        pattern_timestep, or wrap status changes
        
        Parameters
        ----------
        name : str
            Pattern name
        duration : int or float
            Duration (in seconds) of the resampled pattern
        pattern_timestep : int
            Timestep (in seconds) of the resampled pattern
        start_clocktime : int
            Time of day (in seconds from midnight) at which pattern begins
        wrap : bool
            Indicates if the sequence of pattern values repeats
        inplace : bool
            Indicates if the pattern should be modified in place
        
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(duration, (int, float))
        assert isinstance(pattern_timestep, int)
        assert isinstance(start_clocktime, int)
        assert isinstance(wrap, bool)
        assert isinstance(inplace, bool)
        
        # Pattern defined using the current time parameters
        entry = self.get_pattern(name)
        entry_start_clocktime = entry['start_clocktime']

        pattern = self.to_Pattern(name)
        
        # index uses new time parameters
        index = np.arange(start_clocktime, duration, pattern_timestep)
        multipliers = []
        for i in index:
            multipliers.append(pattern.at(i-entry_start_clocktime))
        
        if inplace:
            self.library[name]['start_clocktime'] = start_clocktime
            self.library[name]['pattern_timestep'] = pattern_timestep
            self.library[name]['wrap'] = wrap
            self.library[name]['multipliers'] = list(multipliers)
        
        series = pd.Series(index=index, data=multipliers)
        
        return series
    
    def add_pattern(self, name, entry):
        """
        Add a pattern to the library
        
        Parameters
        ----------
        name : str
            Pattern name
        entry : dict
            Pattern entry which contains the following dictionary keys
            
            * name: Pattern name (string)
            * category: Pattern category (string, optional)
            * description: Pattern description (string, optional)
            * citation: Pattern citation (string, optional)
            * start_clocktime: Time of day (in seconds from midnight) at which pattern begins (integer)
            * pattern_timestep: Pattern timestep in seconds (integer)
            * wrap: Indicates if the sequence of pattern values repeats (True or False)
            * multipliers: Pattern values (list of floats)
        
        Returns
        -------
        pandas Series
        """

        assert isinstance(name, str)
        assert isinstance(entry, dict)
        assert name not in set(self.pattern_name_list)
        
        required_keys = ['name', 
                         'start_clocktime', 
                         'pattern_timestep', 
                         'wrap', 
                         'multipliers']
        assert set(required_keys) <= set(entry.keys())

        self.library[name] = entry
        series = self.to_Series(name)
        
        return series
        
    def add_pulse_pattern(self, name, on_off_sequence, duration=86400, 
                          pattern_timestep=3600, start_clocktime=0, 
                          wrap=True, invert=False, normalize=False):
        """
        Add a pulse pattern to the library using a sequence of on/off times
        
        Pulse patterns can be used to model sudden changes in water demand, 
        for example, from a fire hydrant.
        
        This pattern replicates functionality in Pattern.binary_pattern
        
        Parameters
        ----------
        name : str
            Pattern name
        on_off_sequence : list
            A list of times to turn the pattern on/off (starting with on)
        duration : int or float
            Duration (in seconds) of the resampled pattern
        pattern_timestep : int
            Timestep (in seconds) of the resampled pattern
        start_clocktime : int
            Time of day (in seconds from midnight) at which pattern begins
        wrap : bool
            Indicates if the sequence of pattern values repeats
        invert : bool
            Indicates if the on/off values should be switched
        normalize : bool
            Indicates if the pattern should be normalized
        
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(on_off_sequence, list)
        assert np.all(np.diff(on_off_sequence) > 0) # is monotonically increasing
        assert isinstance(duration, (int, float))
        assert isinstance(pattern_timestep, int)
        assert isinstance(start_clocktime, int)
        assert isinstance(wrap, bool)
        assert isinstance(invert, bool)
        assert isinstance(normalize, bool)

        index = np.arange(start_clocktime, duration, pattern_timestep)
        multipliers = pd.Series(index=index, data=0) # starts off
        switches = 0
        for time in on_off_sequence:
            switches = switches + 1
            position = np.mod(switches,2) # returns 0 or 1
            multipliers.loc[time::] = position

        if invert:
            multipliers = multipliers.max() - multipliers

        if normalize:
            multipliers = multipliers/multipliers.mean()

        entry = {'name': name,
                 'category': None,
                 'description': None,
                 'citation': None,
                 'start_clocktime': start_clocktime,
                 'pattern_timestep': pattern_timestep,
                 'wrap': wrap,
                 'multipliers': list(multipliers)}

        self.add_pattern(name, entry)
        series = self.to_Series(name)
        
        return series

    def add_gaussian_pattern(self, name, mean, std, duration=86400, 
                             pattern_timestep=3600, start_clocktime=0, 
                             wrap=True, invert=False, normalize=False):
        """
        Add a Guassian pattern to the library defined by a mean and standard
        deviation
        
        Gaussian patterns can be used to model water demand that gradually 
        increases to a max water use, followed by gradual decline.
        
        Parameters
        ----------
        name : str
            Pattern name
        mean : int or float
            Mean of the Guassian distribution
        std : int or float
            Standard deviation of the Guassian distribution
        duration : int or float
            Duration (in seconds) of the pattern
        pattern_timestep : int
            Timestep (in seconds) of the pattern
        start_clocktime : int
            Time of day (in seconds from midnight) at which pattern begins
        wrap : bool
            Indicates if the sequence of pattern values repeats
        invert : bool
            Indicates if the on/off values should be switched
        normalize : bool
            Indicates if the pattern should be normalized
            
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(mean, (int, float))
        assert isinstance(std, (int, float))
        assert isinstance(duration, (int, float))
        assert isinstance(pattern_timestep, int)
        assert isinstance(start_clocktime, int)
        assert isinstance(wrap, bool)
        assert isinstance(invert, bool)
        assert isinstance(normalize, bool)
        
        index = np.arange(start_clocktime, duration, pattern_timestep)
        multipliers = scipy.stats.norm.pdf(index, mean, std)

        if invert:
            multipliers = multipliers.max() - multipliers

        if normalize:
            multipliers = multipliers/multipliers.mean()

        entry = {'name': name,
                 'category': None,
                 'description': None,
                 'citation': None,
                 'start_clocktime': start_clocktime,
                 'pattern_timestep': pattern_timestep,
                 'wrap': wrap,
                 'multipliers': list(multipliers)}

        self.add_pattern(name, entry)
        series = self.to_Series(name)
        
        return series

    def add_triangular_pattern(self, name, start, peak, end, duration=86400, 
                               pattern_timestep=3600, start_clocktime=0,
                               wrap=True, invert=False, normalize=False):
        """
        Add a triangular pattern to the library defined by a start time,
        peak time, and end time
        
        Triangular patterns can be used to model water demand that uniformly 
        increases to a max water use, followed by uniform decline.
        
        Parameters
        ----------
        name : str
            Pattern name
        start : int or float
            Start time (in seconds) of the triangular distribution
        peak : int or float
            Peak time (in seconds) of the triangular distribution
        end : int or float
            End time (in seconds) of the triangular distribution
        duration : int or float
            Duration (in seconds) of the pattern
        pattern_timestep : int
            Timestep (in seconds) of the pattern
        start_clocktime : int
            Time of day (in seconds from midnight) at which pattern begins
        wrap : bool
            Indicates if the sequence of pattern values repeats
        invert : bool
            Indicates if the on/off values should be switched
        normalize : bool
            Indicates if the pattern should be normalized
            
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(start, (int, float))
        assert isinstance(peak, (int, float))
        assert isinstance(end, (int, float))
        assert isinstance(duration, (int, float))
        assert isinstance(pattern_timestep, int)
        assert isinstance(start_clocktime, int)
        assert isinstance(wrap, bool)
        assert isinstance(invert, bool)
        assert isinstance(normalize, bool)
        
        loc = start
        scale = end-start
        c = (peak-start)/(end-start)

        index = np.arange(start_clocktime, duration, pattern_timestep)
        multipliers = scipy.stats.triang.pdf(index, c, loc, scale)

        if invert:
            multipliers = multipliers.max() - multipliers

        if normalize:
            multipliers = multipliers/multipliers.mean()

        entry = {'name': name, 
                 'category': None,
                 'description': None,
                 'citation': None,
                 'start_clocktime': start_clocktime,
                 'pattern_timestep': pattern_timestep,
                 'wrap': wrap,
                 'multipliers': list(multipliers)}

        self.add_pattern(name, entry)
        series = self.to_Series(name)
        
        return series
    
    def add_combined_pattern(self, name, patterns_to_combine, combine='Overlap', weights=None, 
                             durations=[86400], pattern_timestep=3600, start_clocktime=0,
                             wrap=True, normalize=False):
        """
        Combine patterns (overlap or sequential) to create a new pattern
        
        Parameters
        ----------
        name : str
            Pattern name
        patterns_to_combine : list of str
            List of pattern names to combine
        combine : str
            Combine method, Overlap or Sequential
        weights: list
            List of weight applied to each pattern.  
            If no weights are provided (None), then the patterns are equally 
            weighted.
        durations : list of ints or floats
            If combine method is Overlap, the list contains only one entry 
            which is the total duration of the pattern (in seconds).
            If combine method is Sequential, the list contains one duration 
            for each pattern to combine (in seconds).
        pattern_timestep : int
            Timestep (in seconds) of the pattern
        start_clocktime : int
            Time of day (in seconds from midnight) at which pattern begins
        wrap : bool
            Indicates if the sequence of pattern values repeats
        normalize : bool
            Indicates if the pattern should be normalized
            
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(patterns_to_combine, list)
        if weights is None:
            weights = [1]*len(patterns_to_combine)
        assert isinstance(weights, list)
        assert len(patterns_to_combine) == len(weights)
        assert isinstance(combine, str)
        assert combine in ['Overlap', 'Sequential']
        assert isinstance(durations, list)
        if combine == 'Overlap':
            assert len(durations) == 1
        else:
            assert len(durations) == len(patterns_to_combine)
        assert isinstance(pattern_timestep, int)
        assert isinstance(start_clocktime, int)
        assert isinstance(wrap, bool)
        assert isinstance(normalize, bool)
        
        t = start_clocktime
        series = {}
        for i, n in enumerate(patterns_to_combine):
            if combine == 'Sequential':
                duration = durations[i]   
            else:
                duration = durations[0]
            weight = weights[i]
            
            entry = self.get_pattern(n)
            #entry_start_clocktime = entry['start_clocktime']
            
            pattern = self.to_Pattern(n)
    
            index = np.arange(t, t+duration, pattern_timestep)
            values = []
            for i in index:
                values.append(pattern.at(i-start_clocktime))
    
            series[n+str(i)] = pd.Series(index=index, data=values*weight)
            
            if combine == 'Sequential':
                t = t + duration
        
        df = pd.DataFrame(index=index, data=series)
        
        if combine == 'Sequential':
            series = pd.concat(series)
            series.droplevel(0)
        else:
            series = df.sum(axis=1)
    
        if normalize:
            series = series/series.mean()
    
        entry = {'name': name, 
                 'category': None,
                 'description': None,
                 'citation': None,
                 'start_clocktime': start_clocktime,
                 'pattern_timestep': pattern_timestep,
                 'wrap': wrap,
                 'multipliers': list(series)}
    
        self.add_pattern(name, entry)
        
        return series

    def to_Pattern(self, name, time_options=None):
        """
        Convert the pattern library entry to a WNTR Pattern
        
        Parameters
        ----------
        name : str
            Pattern name
        time_options : None or tuple 
            Time options (pattern_start, pattern_timestep, pattern_interpolation)
            If None, then time options from the pattern are used (in seconds).
    
        Returns
        -------
        WNTR Pattern object
        """
        
        assert isinstance(name, str)
        assert isinstance(time_options, (TimeOptions, tuple, NoneType))
    
        entry = self.get_pattern(name)
        pattern_timestep = entry['pattern_timestep']
        start_clocktime = entry['start_clocktime']
        
        if time_options is None:
            pattern_start = 0
            pattern_interpolation=True
            time_options = (pattern_start, 
                            pattern_timestep, 
                            pattern_interpolation)
        elif isinstance(time_options, TimeOptions):
            assert time_options.pattern_timestep == pattern_timestep
            assert time_options.start_clocktime == start_clocktime
        elif isinstance(time_options, tuple):
            assert time_options[1] == pattern_timestep
            
        # Note pattern_start is only used by the EpanetSimulator or WNTRSimulator
        multipliers = entry['multipliers']
        wrap = entry['wrap']
        pattern = Pattern(name=name,
                          multipliers=multipliers,
                          time_options=time_options,
                          wrap=wrap)

        return pattern

    def to_Series(self, name, duration=None):
        """
        Convert the pattern library entry to a Pandas Series
        
        Parameters
        ----------
        name : str
            Pattern name
        duration : int, float, or None
            Pattern duration (in seconds).  If None, then the duration from 
            the pattern entry is used.
        
        Returns
        -------
        pandas Series
        """
        
        assert isinstance(name, str)
        assert isinstance(duration, (int, float, NoneType))
        
        entry = self.get_pattern(name)

        start_clocktime = entry['start_clocktime']
        pattern_timestep = entry['pattern_timestep']
        if duration is None:
            multipliers = entry['multipliers']
            duration = start_clocktime + len(multipliers)*pattern_timestep

        pattern = self.to_Pattern(name)

        # Get values at a particular time, can be used to resample
        index = np.arange(start_clocktime, duration, pattern_timestep)
        data = []
        for i in index:
            data.append(pattern.at(i-start_clocktime))
        series = pd.Series(index=index, data=data)

        return series

    def write_json(self, filename):
        """
        Write the library to a JSON file
        
        Parameters
        ----------
        filename : str
            Filename for the library JSON file
        """
        
        assert isinstance(filename, str)
        
        data = []
        for name, entry in self.library.items():
            data.append(entry)
            
        if isinstance(filename, str):
            with open(filename, "w") as fout:
                json.dump(data, fout)

    def plot_patterns(self, names=None, duration=None, ax=None):
        """
        Plot patterns
        
        Parameters
        ----------
        names : list of str
            Pattern names, if None then all patterns are plotted
        duration : int or float
            Pattern duration (in seconds).  If None, then the duration from 
            the pattern entry is used.
        
        Returns
        -------
        matplotlib axes object  
        """
        
        assert isinstance(names, (list, NoneType))
        assert isinstance(duration, (int, float, NoneType))
        
        if names is None:
            names = self.pattern_name_list

        if ax is None:
            fig, ax = plt.subplots()

        for name in names:
            series = self.to_Series(name, duration=duration)
            series.index = series.index/3600
            series.plot(ax=ax, linewidth=1.5, label=name)
            
        ax.set_ylabel('Demand Multiplier')
        ax.set_xlabel('Time (hr)')
        ax.legend()

        return ax



import unittest
from os.path import abspath, dirname, join

import pandas as pd

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestPumpNet3(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

        inp_file = join(ex_datadir, "Net3.inp")
        self.wn = self.wntr.network.WaterNetworkModel(inp_file)
        self.wn.options.energy.global_efficiency = 75  # 75%
        self.wn.options.energy.global_price = 3.61e-8  # $/J; equal to $0.13/kW-h
        self.wn.options.time.hydraulic_timestep = 15*60
        self.wn.options.time.quality_timestep = 15*60
        self.wn.options.time.report_timestep = 15*60
        
    @classmethod
    def tearDownClass(self):
        pass

    def test_energy_report(self):

        sim = self.wntr.sim.EpanetSimulator(self.wn)
        results = sim.run_sim()

        flowrate = results.link["flowrate"].loc[:, self.wn.pump_name_list]
        head = results.node["head"].loc[:, self.wn.node_name_list]
        pump_energy = self.wntr.metrics.pump_energy(flowrate, head, self.wn)

        pump_status = results.link["status"].loc[:, self.wn.pump_name_list]
        utilization = pump_status.mean() * 100

        timestep = self.wn.options.time.report_timestep
        total_pumped_volume_Mgal = (flowrate * timestep * 264.172 / 1e6).sum()
        pump_energy_kWhr = pump_energy / (1000 * 3600)
        total_pump_energy_kWhr = pump_energy_kWhr.sum()
        energy_per_vol = total_pump_energy_kWhr / total_pumped_volume_Mgal  # kW-hr/Mgal

        pump_power_kW = pump_energy_kWhr / (timestep / 3600)
        average_power = {}
        for pump in pump_power_kW.columns:
            average_power[pump] = pump_power_kW.loc[pump_status[pump] == 1, pump].mean()
        average_power = pd.Series(average_power)

        peak_power = pump_power_kW.max()

        pump_cost = self.wntr.metrics.pump_cost(pump_energy, self.wn)
        cost_per_day = pump_cost.sum() / 7  # pump energy (J) * cost ($/J)

        # Compare results to an Energy Report from EPANET GUI (setting global price to $0.13)
        results = pd.DataFrame(index=["10", "335"])
        results["utilization"] = utilization
        results["energy_per_vol"] = energy_per_vol
        results["average_power"] = average_power
        results["peak_power"] = peak_power
        results["cost_per_day"] = cost_per_day

        solution = {
            "utilization": [58.33, 23.60],
            "energy_per_vol": [313.71, 394.29],
            "average_power": [62.05, 309.41],
            "peak_power": [62.76, 310.84],
            "cost_per_day": [112.93, 227.85],
        }
        solution = pd.DataFrame(solution, index=["10", "335"])

        error = abs(results - solution) / solution

        self.assertLess(error.max().max(), 0.01)


if __name__ == "__main__":
    unittest.main()



import unittest
from os.path import abspath, dirname, join

import numpy as np
import wntr

testdir = dirname(abspath(str(__file__)))


class TestEpanetUnits(unittest.TestCase):
    def test_Concentration(self):
        typestring = "Concentration"
        data_expected = 1  # kg/m3
        data = 1000  # mg/L
        for flowunit in range(10):
            self.execute_check_qual(typestring, flowunit, data, data_expected)

    def test_Demand(self):
        data_expected = 1.0  # m/s
        for typestring in ["Demand", "Flow"]:
            for flowunit in range(10):
                if flowunit == 0:
                    data = 35.3146667  # ft3/s
                elif flowunit == 1:
                    data = 15850.3231  # gall/min
                elif flowunit == 2:
                    data = 22.8244653  # million gall/d
                elif flowunit == 3:
                    data = 19.0  # million imperial gall/d
                elif flowunit == 4:
                    data = 70.0456199  # acre-feet/day
                elif flowunit == 5:
                    data = 1000.0  # L/s
                elif flowunit == 6:
                    data = 60000.0  # L/min
                elif flowunit == 7:
                    data = 86.4  # million L/d
                elif flowunit == 8:
                    data = 3600.0  # m3/h
                elif flowunit == 9:
                    data = 86400.0  # m3/d
                self.execute_check(typestring, flowunit, data, data_expected)

    def test_Emitter_Coeff(self):
        data_expected = 1.0  # (m3/s)/sqrt(m)
        typestring = "EmitterCoeff"
        data_expected = 1  # m
        for flowunit in range(10):
            if flowunit == 0:
                data = 35.3146667  # ft3/s
            elif flowunit == 1:
                data = 15850.3231  # gall/min
            elif flowunit == 2:
                data = 22.8244653  # million gall/d
            elif flowunit == 3:
                data = 19.0  # million imperial gall/d
            elif flowunit == 4:
                data = 70.0456199  # acre-feet/day
            elif flowunit == 5:
                data = 1000.0  # L/s
            elif flowunit == 6:
                data = 60000.0  # L/min
            elif flowunit == 7:
                data = 86.4  # million L/d
            elif flowunit == 8:
                data = 3600.0  # m3/h
            elif flowunit == 9:
                data = 86400.0  # m3/d

            if flowunit in [0, 1, 2, 3, 4]:
                data = data/np.sqrt(1.421970206324753)  # flowrate/sqrt(psi)
            else:
                data = data # flowrate/sqrt(m)
                
            self.execute_check(typestring, flowunit, data, data_expected)

    def test_Pipe_Diameter(self):
        typestring = "PipeDiameter"
        data_expected = 1.0  # m
        for flowunit in range(10):
            if flowunit in [0, 1, 2, 3, 4]:
                data = 39.3701  # in
            elif flowunit in [5, 6, 7, 8, 9]:
                data = 1000  # mm
            self.execute_check(typestring, flowunit, data, data_expected)

    def test_Length(self):
        data_expected = 1.0  # m
        for typestring in ["TankDiameter", "Elevation", "HydraulicHead", "Length"]:
            for flowunit in range(10):
                if flowunit in [0, 1, 2, 3, 4]:
                    data = 3.28084  # ft
                elif flowunit in [5, 6, 7, 8, 9]:
                    data = 1  # m
                self.execute_check(typestring, flowunit, data, data_expected)

    def test_Velocity(self):
        typestring = "Velocity"
        data_expected = 1.0  # m/s
        for flowunit in range(10):
            if flowunit in [0, 1, 2, 3, 4]:
                data = 3.28084  # ft/s
            elif flowunit in [5, 6, 7, 8, 9]:
                data = 1  # m/s
            self.execute_check(typestring, flowunit, data, data_expected)

        # test list
        data = np.array([10, 20, 30])  # ft/s
        data_expected = [3.048, 6.096, 9.144]  # m/s
        self.execute_check_list(typestring, 1, data, data_expected)

    def test_Energy(self):
        typestring = "Energy"
        data_expected = 1000000  # J
        data = 0.277777777778  # kW*hr
        for flowunit in range(10):
            self.execute_check(typestring, flowunit, data, data_expected)

    def test_Power(self):
        typestring = "Power"
        data_expected = 1000  # W
        for flowunit in range(10):
            if flowunit in [0, 1, 2, 3, 4]:
                data = 1.34102209  # hp
            elif flowunit in [5, 6, 7, 8, 9]:
                data = 1  # kW
            self.execute_check(typestring, flowunit, data, data_expected)

    def test_Pressure(self):
        typestring = "Pressure"
        data_expected = 1  # m
        for flowunit in range(10):
            if flowunit in [0, 1, 2, 3, 4]:
                data = 1.421970206324753  # psi
            elif flowunit in [5, 6, 7, 8, 9]:
                data = 1  # m
            self.execute_check(typestring, flowunit, data, data_expected)

    def test_Source_Mass_Injection(self):
        typestring = "SourceMassInject"
        data_expected = 1  # kg/s
        data = 6.0e7  # mg/min
        for flowunit in range(10):
            self.execute_check_qual(typestring, flowunit, data, data_expected)

    def test_Volume(self):
        typestring = "Volume"
        data_expected = 1  # m3
        for flowunit in range(10):
            if flowunit in [0, 1, 2, 3, 4]:
                data = 35.3147  # ft3
            elif flowunit in [5, 6, 7, 8, 9]:
                data = 1  # m3
            self.execute_check(typestring, flowunit, data, data_expected)

    def test_Water_Age(self):
        typestring = "WaterAge"
        data_expected = 1  # s
        data = 0.000277778  # hr
        for flowunit in range(10):
            self.execute_check_qual(typestring, flowunit, data, data_expected)

    def execute_check(self, typestring, flowunit, data, data_expected):
        #    data_convert = wntr.utils.convert(typestring, flowunit, data)
        data_convert = wntr.epanet.util.HydParam[typestring]._to_si(
            wntr.epanet.util.FlowUnits(flowunit), data
        )
        self.assertLess(
            abs((data_convert - data_expected) / float(data_expected)), 0.001
        )
        data_convert = wntr.epanet.util.HydParam[typestring]._from_si(
            wntr.epanet.util.FlowUnits(flowunit), data_expected
        )
        self.assertLess(abs((data_convert - data) / data), 0.001)

    def execute_check_list(self, typestring, flowunit, data, data_expected):
        #    data_convert = wntr.utils.convert(typestring, flowunit, data)
        data_convert = wntr.epanet.util.HydParam[typestring]._to_si(
            wntr.epanet.util.FlowUnits(flowunit), data
        )
        data_convert = [round(k, 3) for k in data_convert]
        self.assertListEqual(data_convert, data_expected)

    def execute_check_qual(self, typestring, flowunit, data, data_expected):
        #    data_convert = wntr.utils.convert(typestring, flowunit, data)
        data_convert = wntr.epanet.util.QualParam[typestring]._to_si(
            wntr.epanet.util.FlowUnits(flowunit), data
        )
        self.assertLess(
            abs((data_convert - data_expected) / float(data_expected)), 0.001
        )
        data_convert = wntr.epanet.util.QualParam[typestring]._from_si(
            wntr.epanet.util.FlowUnits(flowunit), data_expected
        )
        self.assertLess(abs((data_convert - data) / data), 0.001)

    def execute_check_qual_list(self, typestring, flowunit, data, data_expected):
        #    data_convert = wntr.utils.convert(typestring, flowunit, data)
        data_convert = wntr.epanet.util.QualParam[typestring]._to_si(
            wntr.epanet.util.FlowUnits(flowunit), data
        )
        data_convert = [round(k, 3) for k in data_convert]
        self.assertListEqual(data_convert, data_expected)


if __name__ == "__main__":
    unittest.main()



import unittest
from os.path import abspath, dirname, join

import pandas as pd
import wntr
from pandas.testing import assert_frame_equal

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestValveLayer(unittest.TestCase):
    @classmethod
    def setUpClass(self):

        inp_file = join(ex_datadir, "Net3.inp")
        self.wn = wntr.network.WaterNetworkModel(inp_file)

    @classmethod
    def tearDownClass(self):
        pass

    def test_valve_layer_random(self):

        valves = wntr.network.generate_valve_layer(self.wn, "random", 40, 123)

        # valves.to_csv('valve_layer_random.csv')
        # ax = wntr.graphics.plot_network(self.wn, node_size=7, valve_layer=valves,
        #                                          title='Random, 40')

        expected = pd.read_csv(
            join(test_datadir, "valve_layer_random.csv"), index_col=0, dtype="object"
        )

        assert_frame_equal(valves, expected)

    def test_valve_layer_strategic(self):

        # Compute the expected number of valves for N-0, N-1, N-2, N-3, N-4
        G = self.wn.to_graph()
        node_degree = pd.Series(dict(G.degree()))
        expected_n_valves = pd.Series(index=[4, 3, 2, 1, 0])
        for N in [4, 3, 2, 1, 0]:
            expected_n_valves[N] = node_degree[node_degree > N].shape[0]

        expected_n_valves = expected_n_valves.cumsum()

        for N in [0, 1, 2, 3, 4]:

            valves = wntr.network.generate_valve_layer(self.wn, "strategic", N, 123)

            # valves.to_csv('valve_layer_stategic_'+str(N)+'.csv')
            # ax = wntr.graphics.plot_network(self.wn, node_size=7, valve_layer=valves,
            #                                          title='Strategic, '+str(N))

            expected_valves = pd.read_csv(
                join(test_datadir, "valve_layer_stategic_" + str(N) + ".csv"),
                index_col=0,
                dtype="object",
            )

            self.assertEqual(valves.shape[0], expected_n_valves[N])
            assert_frame_equal(valves, expected_valves, check_index_type=False)


if __name__ == "__main__":
    unittest.main()



import unittest
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import networkx as nx
import wntr

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestSegmentation(unittest.TestCase):
    @classmethod
    def setUpClass(self):

        inp_file1 = join(test_datadir, "CCWI17-HermanMahmoud.inp")
        self.wn1 = wntr.network.WaterNetworkModel(inp_file1)

        inp_file2 = join(ex_datadir, "Net3.inp")
        self.wn2 = wntr.network.WaterNetworkModel(inp_file2)

    @classmethod
    def tearDownClass(self):
        pass

    def test_segmentation_small(self):
        # test a small network
        G = self.wn1.to_graph()

        valves = [
            ["p1", "n1"],  # valve 0 is on link p1 and protects node n1
            ["p3", "n2"],
            ["p7", "n5"],
            ["p7", "n6"],
            ["p8", "n6"],
        ]
        valves = pd.DataFrame(valves, columns=["link", "node"])

        (
            node_segments,
            link_segments,
            seg_size,
        ) = wntr.metrics.topographic.valve_segments(G, valves)

        max_seg_size = seg_size.sum(axis=1).max()
        self.assertEqual(max_seg_size, 11)
        self.assertEqual(seg_size.shape[0], 4)

    def test_segmentation_random(self):
        # test Net3
        G = self.wn2.to_graph()
        valves = pd.read_csv(
            join(test_datadir, "valve_layer_random.csv"), index_col=0, dtype="object"
        )

        (
            node_segments,
            link_segments,
            seg_size,
        ) = wntr.metrics.topographic.valve_segments(G, valves)
        max_seg_size = seg_size.sum(axis=1).max()
        num_segments = seg_size.shape[0]

        #        node_segments.to_csv('node_segments_random.csv')
        #        link_segments.to_csv('link_segments_random.csv')

        expected_node_segments = pd.read_csv(
            join(test_datadir, "node_segments_random.csv"), index_col=0
        ).squeeze()
        expected_link_segments = pd.read_csv(
            join(test_datadir, "link_segments_random.csv"), index_col=0
        ).squeeze()
        expected_node_segments.astype("int32")
        expected_link_segments.astype("int32")

        #        import matplotlib
        #        cmap = matplotlib.colors.ListedColormap(np.random.rand(num_segments,3))
        #        wntr.graphics.plot_network(self.wn2, node_segments, link_segments, valve_layer=valves,
        #                                   node_cmap=cmap, link_cmap=cmap,
        #                                   node_range=[0.5,num_segments+0.5],
        #                                   link_range=[0.5,num_segments+0.5])

        self.assertListEqual(list(node_segments), list(expected_node_segments))
        self.assertListEqual(list(link_segments), list(expected_link_segments))

        self.assertEqual(max_seg_size, 125)
        self.assertEqual(num_segments, 23)

    def test_segmentation_strategic(self):
        # test Net3
        G = self.wn2.to_graph()
        valves = pd.read_csv(
            join(test_datadir, "valve_layer_stategic_1.csv"),
            index_col=0,
            dtype="object",
        )

        (
            node_segments,
            link_segments,
            seg_size,
        ) = wntr.metrics.topographic.valve_segments(G, valves)
        max_seg_size = seg_size.sum(axis=1).max()
        num_segments = seg_size.shape[0]

        #        node_segments.to_csv('node_segments_strategic.csv')
        #        link_segments.to_csv('link_segments_strategic.csv')

        expected_node_segments = pd.read_csv(
            join(test_datadir, "node_segments_strategic.csv"), index_col=0
        ).squeeze()
        expected_link_segments = pd.read_csv(
            join(test_datadir, "link_segments_strategic.csv"), index_col=0
        ).squeeze()
        expected_node_segments.astype("int32")
        expected_link_segments.astype("int32")

        #        import matplotlib
        #        cmap = matplotlib.colors.ListedColormap(np.random.rand(num_segments,3))
        #        wntr.graphics.plot_network(self.wn2, node_segments, link_segments, valve_layer=valves,
        #                                   node_cmap=cmap, link_cmap=cmap,
        #                                   node_range=[0.5,num_segments+0.5],
        #                                   link_range=[0.5,num_segments+0.5])

        self.assertListEqual(list(node_segments), list(expected_node_segments))
        self.assertListEqual(list(link_segments), list(expected_link_segments))

        self.assertEqual(max_seg_size, 3)
        self.assertEqual(num_segments, 119)

    def test_compare_segmentations(self):
            #compare results from two segmentation algorithms
            G = self.wn2.to_graph()
            
            strategic_valve_layer = wntr.network.generate_valve_layer(
                self.wn2, 'strategic', 1, seed = 123
                )

            (node_segments, 
            link_segments, 
            segment_size) = wntr.metrics.valve_segments(G, strategic_valve_layer)

            (old_node_segments, 
            old_link_segments, 
            old_segment_size) = matrix_valve_segments(G, strategic_valve_layer)

            # basic length checks
            self.assertEqual(len(old_node_segments), len(node_segments))
            self.assertEqual(len(old_link_segments), len(link_segments))
            self.assertEqual(len(old_segment_size), len(segment_size))

            """
            Warning: the following three tests assumes that algorithms output 
            same labels for segment classes, which may not always be the case.
            If this test fails, consider looking into a comparison
            that checks if segment groups are the same, rather
            than exact labelling.
            TODO: use a set-of-sets approach to properly compare
            segmentations without relying on above assumption.
            """
            for link in link_segments.index:
                self.assertEqual(
                    link_segments.loc[link],
                    old_link_segments.loc[link]
                    )

            for node in node_segments.index:
                self.assertEqual(
                    node_segments.loc[node],
                    old_node_segments.loc[node]
                    )

            # check segment sizes
            for k in segment_size.index:
                self.assertTrue(
                    (old_segment_size.loc[k]==segment_size.loc[k]).all()
                    )

def matrix_valve_segments(G, valve_layer):
    """
    Valve segmentation
    This is the original wntr implementation of a segmentation
    method. It is here primarily for use as a "checker" for the
    connected components implementation.

    Parameters
    -----------
    G: networkx MultiDiGraph
        Graph
    valve_layer: pandas DataFrame
        Valve layer, defined by node and link pairs (for example, valve 0 is 
        on link A and protects node B). The valve_layer DataFrame is indexed by
        valve number, with columns named 'node' and 'link'.

    Returns
    -------
    node_segments: pandas Series
       Segment number for each node, indexed by node name
    link_segments: pandas Series
        Segment number for each link, indexed by link name
    segment_size: pandas DataFrame
        Number of nodes and links in each segment. The DataFrame is indexed by 
        segment number, with columns named 'node' and 'link'.
    """
    # Convert the graph to an undirected graph
    uG = G.to_undirected()
    
    # Node and link names
    nodes = list(uG.nodes()) # list of node names
    links = list(uG.edges(keys=True)) # list of tuples with start node, end node, link name
    
    # Append N_ and L_ to node and link names, used in matrices
    matrix_node_names = ['N_'+n for n in nodes]
    matrix_link_names = ['L_'+k for u,v,k in links]

    # Pipe-node connectivity matrix
    AC = nx.incidence_matrix(uG).astype('int').todense().T
    AC = pd.DataFrame(AC, columns=matrix_node_names, index=matrix_link_names, dtype=int)
    
    # Valve-node connectivity matrix
    VC = pd.DataFrame(0, columns=matrix_node_names, index=matrix_link_names)
    for i, row in valve_layer.iterrows():
        VC.at['L_'+row['link'], 'N_'+row['node']] = 1
    
    # Valve deficient matrix (anti-valve matrix)
    VD = AC - VC
    del AC, VC
    
    # Direct connectivity matrix
    NI = pd.DataFrame(np.identity(len(matrix_node_names)),
                      index = matrix_node_names, columns = matrix_node_names)
    LI = pd.DataFrame(np.identity(len(matrix_link_names)),
                      index = matrix_link_names, columns = matrix_link_names)
    DC_left = pd.concat([NI, VD], sort=False)
    DC_right = pd.concat([VD.T, LI], sort=False)
    del LI, VD
    DC = pd.concat([DC_left, DC_right], axis=1, sort=False)
    del DC_left, DC_right
    DC = DC.astype(int)

    # initialization for looping routine
    seg_index = 0
    
    # pre-processing to find isolated elements before looping
    seg_label = {}

    # find links with valves on either end -kb
    for start_node, end_node, link_name in links:
        link_valves = valve_layer[valve_layer['link']==link_name]
        if set(link_valves['node']) >= set([start_node, end_node]):
            seg_index += 1
            seg_label['L_'+link_name] = seg_index

    # find nodes with valves
    for node_name in nodes:
        node_valves = valve_layer[valve_layer['node']==node_name]
        node_links = [k for u,v,k in uG.edges(node_name, keys=True)]
        if set(node_valves['link']) >= set(node_links):
            seg_index += 1
            seg_label['N_'+node_name] = seg_index

    # drop previously found isolated elements before looping
    DC = DC.drop(seg_label.keys())
    DC = DC.drop(seg_label.keys(), axis=1)   
    
    DC_np = DC.to_numpy() # requires Pandas v.0.24.0
    # transpose to align with numpy's row major default
    DC_np = DC_np.transpose()
    # vector of length nodes+links where the ith entry is the segment number of node/link i
    seg_label_DC = np.zeros(shape=(len(DC.index)), dtype=int)

    # Loop over all nodes and links to grow segments
    for i in range(seg_label_DC.shape[0]):
        
        # Only assign a seg_label if node/link doesn't already have one
        if seg_label_DC[i] == 0:
            
            # Advance segment label and assign to node/link, mark as assigned
            seg_index += 1
            seg_label_DC[i] = seg_index
           
            # Initialize segment size
            seg_size = (seg_label_DC == seg_index).sum()
             
            flag = True
            
            #print(i)
    
            # Nodes and links that are part of the segment
            seg = np.where(seg_label_DC == seg_index)[0]
        
            # Connectivity of the segment
            seg_DC = np.zeros(shape=(DC_np.shape[0]), dtype=int)
            seg_DC[seg] = 1
    
            while flag:          
               
                # Potential connectivity of the segment      
                # p_seg_DC = DC_np + seg_DC # this is slow
                connected_to_seg = np.sum(
                    DC_np[:,seg_DC.nonzero()[0]],axis=1
                    ).nonzero()[0]

                # Nodes and links that are connected to the segment
                seg_DC[connected_to_seg] = 1
      
                # Label nodes/links connected to the segment
                seg_label_DC[connected_to_seg] = seg_index
                
                # Find new segment size
                new_seg_size = (seg_label_DC == seg_index).sum()
                    
                # Check for progress
                if seg_size == new_seg_size:
                    flag = False
                else:
                    seg_size = new_seg_size
                # Update seg_DC and DC_np
                seg_DC = np.zeros(seg_DC.shape)
                seg_DC[DC_np[i,:].nonzero()] = 1
                seg_DC[np.sum(DC_np[connected_to_seg,:],axis=0).nonzero()[0]] = 1
                # test = set(DC_np[i,:].nonzero()[0])<=set(np.sum(p_seg_DC[connected_to_seg,:],axis=0).nonzero()[0])
                # print("LOG index comparison: "+str(test))

                # seg_DC = np.clip(seg_DC,0,1)       
                DC_np[connected_to_seg,:] = seg_DC
    
            # print(i, seg_size)
    
    # combine pre-processed and looped results
    seg_labels = list(seg_label.values()) + list(seg_label_DC)
    seg_labels_index = list(seg_label.keys()) + list(DC.index)
    seg_label = pd.Series(seg_labels, index=seg_labels_index, dtype=int)

    # Separate node and link segments
    # remove leading N_ and L_ from node and link names
    node_segments = seg_label[matrix_node_names]
    link_segments = seg_label[matrix_link_names]
    node_segments.index = node_segments.index.str[2::]
    link_segments.index = link_segments.index.str[2::]  
    
    # Extract segment sizes, for nodes and links
    seg_link_sizes = link_segments.value_counts().rename('link')
    seg_node_sizes = node_segments.value_counts().rename('node')
    seg_sizes = pd.concat([seg_link_sizes, seg_node_sizes], axis=1).fillna(0)
    seg_sizes = seg_sizes.astype(int)
    
    return node_segments, link_segments, seg_sizes

if __name__ == "__main__":
    unittest.main()



import unittest
from os.path import abspath, dirname, join
import sys, platform
import wntr

if 'darwin' in sys.platform.lower() and 'arm' in platform.platform().lower():
    skip_v2_tests_on_arm = True
else:
    skip_v2_tests_on_arm = False

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "..", "..", "examples", "networks")

"""
Compare results to EPANET GUI using Net3 and Source Quality at
121          SETPOINT      100000  mg/L
121          FLOWPACED     100000  mg/L
121          MASS          6000000000  mg/min
River        CONCEN        100000   mg/L
"""


class TestWaterQualitySimulations(unittest.TestCase):
    def test_setpoint_waterquality_simulation(self):
        inp_file = join(datadir, "Net3.inp")
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "CHEMICAL"
        wn.add_pattern(
            "NewPattern", [1]
        )  # start_time=0, end_time=wn.options.time.duration)
        wn.add_source("Source1", "121", "SETPOINT", 100, "NewPattern")
        # WQ = wntr.scenario.Waterquality('CHEM', ['121'], 'SETPOINT', 100, 0, -1)

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        expected = 91661.72 * (1e-6 / 0.001)  # Node '159' at hour 6
        # print(results.node)
        error = abs(
            (results.node["quality"].loc[6 * 3600, "159"] - expected) / expected
        )
        self.assertLess(error, 0.0001)  # 0.01% error

    def test_flowpaced_waterquality_simulation(self):
        inp_file = join(datadir, "Net3.inp")
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "CHEMICAL"
        wn.add_pattern(
            "NewPattern", [1]
        )  # start_time=0, end_time=wn.options.time.duration)
        wn.add_source("Source1", "121", "FLOWPACED", 100, "NewPattern")
        # WQ = wntr.scenario.Waterquality('CHEM', ['121'], 'FLOWPACED', 100, 0, -1)

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        expected = 92246.55 * (1e-6 / 0.001)  # Node '159' at hour 6
        error = abs(
            (results.node["quality"].loc[6 * 3600, "159"] - expected) / expected
        )
        self.assertLess(error, 0.0001)  # 0.01% error

    def test_mass_waterquality_simulation(self):
        inp_file = join(datadir, "Net3.inp")
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "CHEMICAL"
        wn.add_pattern(
            "NewPattern", [1.0]
        )  # start_time=0, end_time=wn.options.time.duration)
        wn.add_source("Source1", "121", "MASS", 100, "NewPattern")
        # WQ = wntr.scenario.Waterquality('CHEM', ['121'], 'MASS', 100, 0, -1)

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        expected = 217903.60 * (1e-6 / 0.001)  # Node '159' at hour 6
        error = abs(
            (results.node["quality"].loc[6 * 3600, "159"] - expected) / expected
        )
        self.assertLess(error, 0.0001)  # 0.01% error

    def test_conc_waterquality_simulation(self):
        inp_file = join(datadir, "Net3.inp")
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "CHEMICAL"
        wn.add_pattern(
            "NewPattern", [1.0]
        )  # start_time=0, end_time=wn.options.time.duration)
        wn.add_source("Source1", "River", "CONCEN", 100, "NewPattern")
        # WQ = wntr.scenario.Waterquality('CHEM', ['River'], 'CONCEN', 100, 0, -1)

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        expected = 91661.72 * (1e-6 / 0.001)  # Node '159' at hour 6
        error = abs(
            (results.node["quality"].loc[6 * 3600, "159"] - expected) / expected
        )
        self.assertLess(error, 0.0001)  # 0.01% error

    def test_age_waterquality_simulation(self):

        inp_file = join(datadir, "Net3.inp")
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "AGE"
        # WQ = wntr.scenario.Waterquality('AGE')

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        # WARNING: This does NOT match the EPANET Windows results - it does match
        # the epanet linux binary
        expected = 3.652 * 3600  # Node '159' at hour 6
        error = abs(
            (results.node["quality"].loc[6 * 3600, "159"] - expected) / expected
        )
        # print([expected, results.node['quality'].loc[6*3600, '159']])
        self.assertLess(error, 0.001)  # 0.01% error

    def test_trace_waterquality_simulation(self):
        inp_file = join(datadir, "Net3.inp")
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "TRACE"
        wn.options.quality.trace_node = "121"
        # WQ = wntr.scenario.Waterquality('TRACE', ['121'])

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)
        # print(results.node.keys())
        expected = 91.66  # Node '159' at hour 6
        error = abs(
            float(results.node["quality"].loc[6 * 3600, "159"] - expected)
            / float(expected)
        )
        self.assertLess(error, 0.0001)  # 0.01% error


if __name__ == "__main__":
    unittest.main()




from __future__ import division

import unittest
from os.path import abspath, dirname, join

import pandas as pd
import wntr
from pandas.testing import assert_frame_equal, assert_series_equal

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
net3dir = join(testdir, "..", "..", "examples", "networks")


class TestMetricsDemand(unittest.TestCase):
    def test_expected_demand_net3_node101(self):
        inp_file = join(net3dir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        expected_demand = wntr.metrics.hydraulic.expected_demand(wn)
        ave_expected_demand = wntr.metrics.hydraulic.average_expected_demand(wn)

        ex_demand_101 = expected_demand["101"]
        ave_ex_demand_101 = ave_expected_demand["101"]

        expected = 0.012813608
        error = abs((ex_demand_101.mean() - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

        error = abs((ave_ex_demand_101 - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

    def test_expected_demand_category(self):
        inp_file = join(net3dir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        node = wn.get_node("123")
        node.demand_timeseries_list[0].category = "A"

        ave_expected_demand = wntr.metrics.hydraulic.average_expected_demand(wn)
        ave_expected_demand_A = wntr.metrics.hydraulic.average_expected_demand(
            wn, category="A"
        )

        error = abs(ave_expected_demand["123"] - ave_expected_demand_A["123"])
        self.assertLess(error, 1e-7)

        error = abs(
            ave_expected_demand_A.sum() - ave_expected_demand_A["123"]
        )  # all other entries are 0
        self.assertLess(error, 1e-7)

    def test_wsa(self):

        expected_demand = pd.DataFrame(
            data=[[12, 2], [3, 4], [5, 10]], columns=["A", "B"], index=[0, 1, 2]
        )
        demand = pd.DataFrame(
            data=[[5, 2], [3, 2], [3, 4]], columns=["A", "B"], index=[0, 1, 2]
        )

        # WSA at each junction and time
        wsa = wntr.metrics.hydraulic.water_service_availability(expected_demand, demand)
        expected = pd.DataFrame(
            data=[[5 / 12, 2 / 2], [3 / 3, 2 / 4], [3 / 5, 4 / 10]],
            columns=["A", "B"],
            index=[0, 1, 2],
        )
        assert_frame_equal(wsa, expected, check_dtype=False)

        # WSA at each junction
        wsa = wntr.metrics.hydraulic.water_service_availability(
            expected_demand.sum(axis=0), demand.sum(axis=0)
        )
        expected = pd.Series(
            data=[(5 + 3 + 3) / (12 + 3 + 5), (2 + 2 + 4) / (2 + 4 + 10)],
            index=["A", "B"],
        )
        assert_series_equal(wsa, expected, check_dtype=False)

        # WSA at each time
        wsa = wntr.metrics.hydraulic.water_service_availability(
            expected_demand.sum(axis=1), demand.sum(axis=1)
        )
        expected = pd.Series(
            data=[(5 + 2) / (12 + 2), (3 + 2) / (3 + 4), (3 + 4) / (5 + 10)],
            index=[0, 1, 2],
        )
        assert_series_equal(wsa, expected, check_dtype=False)

class TestTankCapacity(unittest.TestCase):
    
    @classmethod
    def setUpClass(self):
        inp_file = join(net3dir, "Net3.inp")
        self.wn = wntr.network.WaterNetworkModel(inp_file)
        self.wn.options.time.duration = 2*24*3600 # 2 days
        sim = wntr.sim.WNTRSimulator(self.wn)
        self.results = sim.run_sim()
    
    @classmethod
    def tearDownClass(self):
        pass
    
    def test_tank_capacity(self):
        
        pressure = self.results.node["pressure"].loc[:,self.wn.tank_name_list]
        tank_capacity = wntr.metrics.tank_capacity(pressure, self.wn)

        self.assertLess(tank_capacity.max().max(), 1)
        self.assertGreater(tank_capacity.min().min(), 0.4) # for this example, tanks capcity is > 0.4
    
    
if __name__ == "__main__":
    unittest.main()
